#define DEBUG(n,x)
#include "VisMonitoring/DQMServer/interface/VisDQMImageServer.h"
#include "VisMonitoring/DQMServer/interface/VisDQMRenderDaemon.h"
#include "VisMonitoring/DQMServer/interface/VisDQMTools.h"
#include "TObjString.h"
#include "TDirectory.h"
#include "TFile.h"
#include "TKey.h"
#include "TH1.h"
#include <sys/time.h>

using namespace lat;

class VisDQMDataImageServer : public DQMImplNet<VisDQMObject>,
			      public VisDQMImageServer
{
public:
  static const uint32_t DQM_MSG_GET_IMAGE_DATA	= 4;
  static const uint32_t DQM_REPLY_IMAGE_DATA	= 105;

  VisDQMDataImageServer(const ServerInfo &info)
    : DQMImplNet<VisDQMObject>(info.argv[0]),
      VisDQMImageServer(info)
    {
      logme()
	<< "INFO: DQM data image server started"
	<< " in path '" << info_.statedir
	<< "' and at request port " << info_.port
	<< ", black list in " << info_.blacklistFile << std::endl;

      // Initialise plugins now.
      initPlugins();

      // Establish the server side.
      debug(info_.debug);
      delay(100);
      startLocalServer(info_.port);
    }

  ~VisDQMDataImageServer(void)
    {
      shutPlugins();
    }

protected:
  // Return the logging prefix.
  virtual std::ostream &
  logme(void)
    {
      return DQMNet::logme();
    }

  // Check if the server should stop.
  virtual bool
  shouldStop(void)
    {
      return info_.stopfile.exists();
    }

  // Purge dead objects.  This cannot ever happen for this server
  // because we are never connected to any network data source.
  virtual void
  purgeDeadObject(VisDQMObject &o)
    {
      ASSERT(false);
    }
  
  // Release objects from wait.  This cannot ever happen because this
  // server is never connected to a network data source which it would
  // have to wait for.
  void
  releaseFromWait(Bucket *msg, WaitObject &w, Object *o)
    {
      ASSERT(false);
    }

  // Web server facing message handling.  Respond to image requests.
  bool
  onMessage(Bucket *msg, Peer *p, unsigned char *data, size_t len)
    {
      // Decode and process this message.
      uint32_t type;
      memcpy (&type, data + sizeof(uint32_t), sizeof (type));
      if (type == DQM_MSG_GET_IMAGE_DATA)
      {
	if (debug_)
	  logme()
	    << "DEBUG: received message 'GET IMAGE DATA' from peer "
	    << p->peeraddr << std::endl;

	if (len < 6*sizeof(uint32_t))
	{
	  logme()
	    << "ERROR: corrupt 'GET IMAGE DATA' message of length " << len
	    << " from peer " << p->peeraddr << std::endl;
	  return false;
	}

	uint32_t namelen;
	uint32_t speclen;
	uint32_t infolen;
	uint32_t objlen;
	memcpy(&namelen, data + 2*sizeof(uint32_t), sizeof(namelen));
	memcpy(&speclen, data + 3*sizeof(uint32_t), sizeof(speclen));
	memcpy(&infolen, data + 4*sizeof(uint32_t), sizeof(infolen));
	memcpy(&objlen, data + 5*sizeof(uint32_t), sizeof(objlen));
	if (len != 6*sizeof(uint32_t) + namelen + speclen + infolen + objlen)
	{
	  logme()
	    << "ERROR: corrupt 'GET IMAGE DATA' message of length " << len
	    << " from peer " << p->peeraddr
	    << ", expected length " << (6*sizeof(uint32_t))
	    << " + " << namelen << " + " << speclen
	    << " + " << infolen << " + " << objlen
	    << std::endl;
	  return false;
	}

	std::string name((char *) data + 6*sizeof(uint32_t), namelen);
	std::string spec((char *) data + 6*sizeof(uint32_t) + namelen, speclen);
	std::string info((char *) data + 6*sizeof(uint32_t) + namelen + speclen, infolen);
	std::string obj ((char *) data + 6*sizeof(uint32_t) + namelen + speclen + infolen, objlen);

        // Give myself 30 seconds, then die.  ROOT sometimes gets stuck
        // here in an infinite loop when closing files, presumably due
        // to an occasional rampage around in memory while rendering.
        struct itimerval timer;
        timer.it_value.tv_sec = 30;
        timer.it_value.tv_usec = 0;
        timer.it_interval.tv_sec = 0;
        timer.it_interval.tv_usec = 0;
        setitimer(ITIMER_VIRTUAL, &timer, 0);

	// FIXME: Image caching? We lose the object each time around here.
	VisDQMObject o;
	o.version = 0;
	o.name = name;
	o.object = 0; // defaults to missing in action
	o.reference = 0;
	o.flags = 0;

	if (! infolen && objlen)
	  // Draw scalar only if the string is non-empty.
	  // Empty string means "missing in action".
	  o.object = new TObjString(obj.c_str());
	else
	{
	  // FIXME: qreports.
	  loadStreamerInfo(info);
	  o.rawdata.insert(o.rawdata.end(),
			   (unsigned char *) &obj[0],
			   (unsigned char *) &obj[0] + objlen);
	  reconstructObject(o);
	}

        // Let timer go now, we made it through the file management,
	// and the rendering has its own internal suicide timer.
        timer.it_value.tv_sec = 0;
        timer.it_value.tv_usec = 0;
        timer.it_interval.tv_sec = 0;
        timer.it_interval.tv_usec = 0;
        setitimer(ITIMER_VIRTUAL, &timer, 0);

	// Now render.
	sendImageToPeer(msg, p, o, spec);

	// Release the objects.
	delete o.object;
	delete o.reference;
	return true;
      }
      else
      {
	logme()
	  << "ERROR: unrecognised message of length " << len
	  << " and type " << type << " from peer " << p->peeraddr
	  << std::endl;
	return false;
      }
  }

  // Send an image of an object to a peer.
  virtual void
  sendImageToPeer(Bucket *msg, Peer *p,
		  VisDQMObject &o,
		  const std::string &spec)
    {
      uint32_t datalen = 0;
      unsigned char *data = 0;

      // Mark the object referenced.
      o.lastreq = Time::current();

      // Validate the image request.
      const char *error = 0;
      VisDQMImgInfo *img = parseImageSpec(o, spec, error);
      if (! img)
      {
	logme()
	  << "ERROR: invalid image specification '"
	  << spec << "' for object '" << o.name
	  << "': " << error << std::endl;
	return;
      }

      // The object is known and the image specification is valid.  Check
      // if we have an already rendered image, and if so, use it as such.
      if (img->version == o.version)
      {
	datalen = img->imgdata.size();
	data = (datalen ? &img->imgdata[0] : 0);
      }

      // We have an object, but no valid image yet.  Render it now.
      else
      {
	doRender(o, *img);
	datalen = img->imgdata.size();
	data = (datalen ? &img->imgdata[0] : 0);
      }

      // Build the response.
      uint32_t words[2];
      words[0] = sizeof(words) + datalen;
      words[1] = DQM_REPLY_IMAGE_DATA;

      msg->data.reserve(msg->data.size() + words[0]);
      copydata(msg, &words[0], sizeof(words));
      if (datalen)
	copydata(msg, data, datalen);
    }


  // Parse an arbitrary string, for extracting the file name.
  bool
  parseString(const char *&p, const char *name, size_t len, std::string &value)
    {
      if (! strncmp(p, name, len))
      {
	p += len;
	len = 0;
	while (p[len] && p[len] != ';')
	  ++len;
	value.insert(value.end(), p, p+len);
	p += len;
	return true;
      }
      return false;
    }
};

class VisDQMDataDaemon : public VisDQMRenderDaemon
{
public:
  VisDQMDataDaemon(int argc, char **argv)
    : VisDQMRenderDaemon(argc, argv)
    {}

protected:
  virtual void
  runServer(const VisDQMImageServer::ServerInfo &info)
    {
      VisDQMDataImageServer server(info);
      server.run();
    }
};


int main(int argc, char **argv)
{
  VisDQMDataDaemon d(argc, argv);
  return d.execute();
}
