// Shut up warning about _POSIX_C_SOURCE mismatch, caused by a
// harmless conflict between _GNU_SOURCE and python header files.
#define NDEBUG 1
#include <features.h>
#if _POSIX_C_SOURCE != 200112L
# undef _POSIX_C_SOURCE
# define _POSIX_C_SOURCE 200112L
#endif
#define DEBUG(n,x)

#include "VisMonitoring/DQMServer/interface/Objects.h"
#include "VisMonitoring/DQMServer/interface/NatSort.h"
#include "VisMonitoring/DQMServer/interface/StringAtom.h"
#include "VisMonitoring/DQMServer/interface/VisDQMIndex.h"
#include "VisMonitoring/DQMServer/interface/VisDQMFile.h"
#include "VisMonitoring/DQMServer/interface/VisDQMCache.h"
#include "VisMonitoring/DQMServer/interface/VisDQMTools.h"
#include "DQMServices/Core/interface/DQMNet.h"

#include "classlib/utils/DebugAids.h"
#include "classlib/utils/TimeInfo.h"
#include "classlib/utils/Signal.h"
#include "classlib/utils/Regexp.h"
#include "classlib/utils/RegexpMatch.h"
#include "classlib/utils/StringOps.h"
#include "classlib/utils/StringFormat.h"
#include "classlib/iobase/Filename.h"
#include "classlib/iobase/IOStatus.h"
#include "classlib/iobase/File.h"

#include "boost/shared_ptr.hpp"
#include "boost/python.hpp"
#include "boost/python/stl_iterator.hpp"

#include "TObjString.h"
#include "TROOT.h"

#include <set>
#include <stdexcept>
#include <ext/hash_map>
#include <dlfcn.h>

using namespace lat;
using namespace boost;
namespace ext = __gnu_cxx;
namespace py = boost::python;
typedef std::pair<std::string, std::string> StringPair;
typedef std::set<StringAtom> StringAtomSet;
typedef std::vector<StringAtom> StringAtomList;
typedef std::map<StringAtom, StringAtomList> StringAtomParentList;

static pthread_rwlock_t parentslock;
static StringAtomParentList parents;
static StringAtomTree stree(1024*1024);
static Regexp RX_THOUSANDS("(\\d)(\\d{3}($|\\D))");
static Regexp RX_OUTER_WHITE("^\\s+|\\s+$");
static Regexp RX_CMSSW_VERSION("CMSSW(?:_[0-9])+(?:_pre[0-9]+)?");
static const uint32_t NET_HAS_ALARM = (DQMNet::DQM_FLAG_REPORT_ERROR
				       | DQMNet::DQM_FLAG_REPORT_WARNING
				       | DQMNet::DQM_FLAG_REPORT_OTHER);

namespace __gnu_cxx
{
  template<> struct hash<StringAtom>
  {
    size_t operator()(const StringAtom &s) const
    { return s.index(); }
  };

  template<> struct hash<std::string>
  {
    size_t operator()(const std::string &s) const
    { return __stl_hash_string(s.c_str()); }
  };
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class PyReleaseInterpreterLock
{
  PyThreadState *save_;
public:
  PyReleaseInterpreterLock(void) { save_ = PyEval_SaveThread(); }
  ~PyReleaseInterpreterLock(void) { PyEval_RestoreThread(save_); }
};

class Lock
{
  pthread_mutex_t *lock_;
public:
  Lock(pthread_mutex_t *lock) { pthread_mutex_lock(lock_ = lock); }
  ~Lock(void)                 { pthread_mutex_unlock(lock_); }
};

class RDLock
{
  pthread_rwlock_t *lock_;
public:
  RDLock(pthread_rwlock_t *lock) { pthread_rwlock_rdlock(lock_ = lock); }
  ~RDLock(void)                  { pthread_rwlock_unlock(lock_); }
};

class WRLock
{
  pthread_rwlock_t *lock_;
public:
  WRLock(pthread_rwlock_t *lock) { pthread_rwlock_wrlock(lock_ = lock); }
  ~WRLock(void)                  { pthread_rwlock_unlock(lock_); }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
struct VisDQMAxisOptions;
struct VisDQMDrawOptions;
struct VisDQMLayoutItem;
struct VisDQMLayoutRow;
struct VisDQMItem;
struct VisDQMStatus;
struct VisDQMShownItem;
struct VisDQMShownRow;
struct VisDQMStatus;
struct VisDQMEventNum;
struct VisDQMSample;
class VisDQMSource;

typedef std::vector< shared_ptr<VisDQMLayoutItem> >	     VisDQMLayoutItems;
typedef std::vector< shared_ptr<VisDQMLayoutRow> >	     VisDQMLayoutRows;
typedef std::vector< shared_ptr<VisDQMShownItem> >	     VisDQMShownItems;
typedef std::vector< shared_ptr<VisDQMShownRow> >	     VisDQMShownRows;
typedef ext::hash_map<StringAtom, shared_ptr<VisDQMItem> >   VisDQMItems;
typedef ext::hash_map<StringAtom, VisDQMStatus>		     VisDQMStatusMap;
typedef std::map<StringAtom, VisDQMDrawOptions>		     VisDQMDrawOptionMap;
typedef std::list<VisDQMEventNum>		 	     VisDQMEventNumList;
typedef std::vector<VisDQMSample>		 	     VisDQMSamples;

struct VisDQMAxisOptions
{
  std::string			min;
  std::string			max;
  std::string			type;
};

struct VisDQMDrawOptions
{
  VisDQMAxisOptions		xaxis;
  VisDQMAxisOptions		yaxis;
  VisDQMAxisOptions		zaxis;
  std::string			drawopts;
  std::string			withref;
};

struct VisDQMLayoutItem
{
  StringAtom			path;
  std::string			desc;
  VisDQMDrawOptions		drawopts;
};

struct VisDQMLayoutRow
{
  VisDQMLayoutItems		columns;
};

struct VisDQMItem
{
  uint32_t			flags;
  uint64_t			version;
  StringAtom			name;
  StringAtomList		parents;
  std::string			data;
  VisDQMSource		 	*plotter;
  shared_ptr<VisDQMLayoutRows>	layout;
};

struct VisDQMShownItem
{
  uint64_t			version;
  int				nalarm;
  int				nlive;
  StringAtom			name;
  std::string			desc;
  VisDQMSource			*plotter;
  const char			*type;
  VisDQMDrawOptions		drawopts;
};

struct VisDQMShownRow
{
  VisDQMShownItems		columns;
};

struct VisDQMStatus
{
  int				nleaves;
  int				nalarm;
  int				nlive;
};

struct VisDQMEventNum
{
  std::string			subsystem;
  long				runnr;
  long				luminr;
  long				eventnr;
  long				runstart;
};

enum VisDQMSampleType
{
  SAMPLE_LIVE,
  SAMPLE_ONLINE_DATA,
  SAMPLE_OFFLINE_DATA,
  SAMPLE_OFFLINE_RELVAL,
  SAMPLE_OFFLINE_MC
};

struct VisDQMSample
{
  VisDQMSampleType		type;
  long				runnr;
  std::string			dataset;
  std::string			version;
};

enum VisDQMObjType
{
  OBJ_SCALAR,
  OBJ_ROOTOBJ
};

static const char *sampleTypeLabel[] =
{
  "live",
  "online_data",
  "offline_data",
  "offline_relval",
  "offline_mc"
};

static const char *objTypeLabel[] =
{
  "scalar",
  "rootobj"
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
static std::ostream &
lograw(const char *label)
{
  return std::cout
    << Time::current().format(true, "[%d/%b/%Y:%H:%M:%S]")
    << label;
}

static std::ostream &
loginfo(void)
{
  return lograw("  INFO: ");
}

static std::ostream &
logerr(void)
{
  return lograw("  ERROR: ");
}

static std::ostream &
logwarn(void)
{
  return lograw("  WARNING: ");
}

static std::string
thousands(const std::string &arg)
{
  std::string result;
  std::string tmp(arg);
  result.reserve(arg.size() + arg.size()/3 + 1);

  while (true)
  {
    result = StringOps::replace(tmp, RX_THOUSANDS, "\\1'\\2");
    if (result == tmp)
      break;

    tmp = result;
  }

  return result;
}

static std::string
stringToJSON(const std::string &x, bool emptyIsNone = false)
{
  if (emptyIsNone && x.empty())
    return "None";

  char buf[8];
  std::string result;
  result.reserve(2 * x.size());
  result += '"';
  for (size_t i = 0, e = x.size(); i < e; ++i)
  {
    unsigned int ch = (unsigned char) x[i];
    if (ch == '\\')
    {
      result += '\\';
      result += '\\';
    }
    else if (ch == '"')
    {
      result += '\\';
      result += (char) ch;
    }
    else if (ch == '\n')
    {
      result += '\\';
      result += 'n';
    }
    else if (ch == '\r')
    {
      result += '\\';
      result += 'r';
    }
    else if (ch == '\t')
    {
      result += '\\';
      result += 't';
    }
    else if (! isprint(ch))
    {
      sprintf(buf, "\\u%04x", ch);
      result += buf;
    }
    else
      result += (char) ch;
  }
  
  result += '"';
  return result;
}

static shared_ptr<Regexp>
rx(const std::string &match, int options = 0)
{
  shared_ptr<Regexp> r;
  if (! match.empty())
  {
    r.reset(new Regexp(match, options));
    if (! r->valid())
      throw std::runtime_error(StringFormat
			       ("Invalid regular expression syntax at"
				" character %2 of '%1': %3")
			       .arg(match)
			       .arg(r->errorOffset())
			       .arg(r->errorMessage()));
    r->study();
  }
  return r;
}

static void
makerx(const std::string &rxstr,
       shared_ptr<Regexp> &rxobj,
       std::string &rxerr,
       int options = 0)
{
  try
  {
    if (! rxstr.empty())
      rxobj = rx(rxstr, options);
  }
  catch (Error &e)
  {
    rxerr = e.explainSelf();
  }
  catch (std::exception &e)
  {
    rxerr = e.what();
  }
}

static void
buildParentNames(StringAtomList &to, const StringAtom &from)
{
  pthread_rwlock_rdlock(&parentslock);
  StringAtomParentList::iterator pos = parents.find(from);
  if (pos == parents.end())
  {
    pthread_rwlock_unlock(&parentslock);
    pthread_rwlock_wrlock(&parentslock);
    pos = parents.find(from);
  }

  if (pos == parents.end())
  {
    std::string parentstr;
    const std::string &fromstr = from.string();
    to.reserve(std::count(fromstr.begin(), fromstr.end(), '/')+1);
    parentstr.reserve(fromstr.size());
  
    for (size_t slash = 0; slash < fromstr.size();
         slash = fromstr.find('/', ++slash))
    {
      parentstr.assign(fromstr, 0, slash);
      to.push_back(StringAtom(&stree, parentstr));
    }
    parents[from] = to;
  }
  else
    to = pos->second;

  pthread_rwlock_unlock(&parentslock);
}

static bool
isScalarType(uint32_t flags)
{
  uint32_t type = (flags & VisDQMIndex::SUMMARY_PROP_TYPE_MASK);
  return (type >= VisDQMIndex::SUMMARY_PROP_TYPE_INT
	  && type <= VisDQMIndex::SUMMARY_PROP_TYPE_STRING);
}

static bool
isIntegerType(uint32_t flags)
{
  uint32_t type = (flags & VisDQMIndex::SUMMARY_PROP_TYPE_MASK);
  return type == VisDQMIndex::SUMMARY_PROP_TYPE_INT;
}

static bool
isRealType(uint32_t flags)
{
  uint32_t type = (flags & VisDQMIndex::SUMMARY_PROP_TYPE_MASK);
  return type == VisDQMIndex::SUMMARY_PROP_TYPE_REAL;
}

static void
translateDrawOptions(py::dict opts, VisDQMDrawOptions &o)
{
  py::stl_input_iterator<std::string> ki(opts), ke;
  for ( ; ki != ke; ++ki)
  {
    std::string key(*ki);
    std::string value(py::extract<std::string>(opts.get(key)));
    if (key == "withref")        o.withref = value;
    else if (key == "drawopts")  o.drawopts = value;
    else if (key == "xmin")      o.xaxis.min = value;
    else if (key == "xmax")      o.xaxis.max = value;
    else if (key == "xtype")     o.xaxis.type = value;
    else if (key == "ymin")      o.yaxis.min = value;
    else if (key == "ymax")      o.yaxis.max = value;
    else if (key == "ytype")     o.yaxis.type = value;
    else if (key == "zmin")      o.zaxis.min = value;
    else if (key == "zmax")      o.zaxis.max = value;
    else if (key == "ztype")     o.zaxis.type = value;
  }
}

static std::string
formatStartTime(long curStartTime)
{
  Time now = Time::current();
  Time runstart = (curStartTime <= 0 ? now : Time(curStartTime, 0));
  tm nowtm = now.utc();
  tm runtm = runstart.utc();
  bool sameyear = nowtm.tm_year == runtm.tm_year;
  bool samemonth = (sameyear && nowtm.tm_mon == runtm.tm_mon);
  bool sameday = (samemonth && nowtm.tm_mday == runtm.tm_mday);
  const char *fmt = (curStartTime <= 0 ? "(Not recorded)"
		     : sameday ? "Today %H:%M"
		     : samemonth ? "%a %d, %H:%M"
		     : sameyear ? "%a %b %d, %H:%M"
		     : "%a %b %d '%y, %H:%M");
  return runstart.format(false, fmt);
}

static void
getEventInfoNum(const std::string &name,
		const char *data,
		VisDQMEventNumList &eventnums)
{
  long value = -1;
  double real;
  size_t len;

  if ((len = 15) > 0
      && name.size() > len
      && name.compare(name.size() - len, len, "/EventInfo/iRun") == 0)
    value = strtol(data, 0, 10);
  else if ((len = 23) > 0
	   && name.size() > len
	   && name.compare(name.size() - len, len, "/EventInfo/iLumiSection") == 0)
    value = strtol(data, 0, 10);
  else if ((len = 17)
	   && name.size() > len
	   && name.compare(name.size() - len, len, "/EventInfo/iEvent") == 0)
    value = strtol(data, 0, 10);
  else if ((len = 28)
	   && name.size() > len
	   && name.compare(name.size() - len, len, "/EventInfo/runStartTimeStamp") == 0)
    if ((real = atof(data)) > 3600)
      value = long(real+0.5);

  if (value >= 0)
  {
    bool found = false;
    std::string subsys(name, 0, name.size() - len);
    VisDQMEventNumList::iterator li = eventnums.begin();
    VisDQMEventNumList::iterator le = eventnums.end();
    for ( ; li != le; ++li)
      if ((found = (li->subsystem == subsys)))
	break;

    VisDQMEventNum *evnum;
    if (! found)
    {
      eventnums.push_front(VisDQMEventNum());
      evnum = &eventnums.front();
      evnum->subsystem = subsys;
      evnum->runnr = -1;
      evnum->luminr = -1;
      evnum->eventnr = -1;
      evnum->runstart = -1;
    }
    else
      evnum = &*li;

    if (len == 15)
      evnum->runnr = value;
    else if (len == 23)
      evnum->luminr = value;
    else if (len == 17)
      evnum->eventnr = value;
    else if (len == 28)
      evnum->runstart = value;
  }
}

// Pick greatest run/lumi/event number combo seen.
static void
setEventInfoNums(const VisDQMEventNumList &eventnums, VisDQMEventNum &current)
{
  VisDQMEventNumList::const_iterator li = eventnums.begin();
  VisDQMEventNumList::const_iterator le = eventnums.end();
  for ( ; li != le; ++li)
  {
    if (li->runnr > current.runnr)
      current = *li;

    if (li->runnr == current.runnr && li->luminr > current.luminr)
      current.luminr = li->luminr;

    if (li->runnr == current.runnr && li->eventnr > current.eventnr)
      current.eventnr = li->eventnr;

    if (li->runnr == current.runnr
	&& li->runstart >= 0
	&& li->runstart < current.runstart)
      current.runstart = li->runstart;
  }
}

inline bool
orderSamplesByDataset(const VisDQMSample &a, const VisDQMSample &b)
{
  // - Order by sample type (ascending).
  if (a.type < b.type)
    return true;
  else if (a.type > b.type)
    return false;

  // - Order by dataset (descending, natural string comparison).
  if (natcmpstr(a.dataset, b.dataset))
    return true;
  else if (natcmpstr(b.dataset, a.dataset))
    return false;

  // - Order by run (descending).
  if (a.runnr > b.runnr)
    return true;
  else if (a.runnr < b.runnr)
    return false;

  // - Order by version (descending, natural string comparison).
  if (natcmpstr(b.version, a.version))
    return true;
  else if (natcmpstr(a.version, b.version))
    return false;

  // Should not come here, but give deterministic order.
  return &a < &b;
}

inline bool
orderSamplesByRun(const VisDQMSample &a, const VisDQMSample &b)
{
  // - Order by sample type (ascending).
  if (a.type < b.type)
    return true;
  else if (a.type > b.type)
    return false;

  // - Order by run (descending).
  if (a.runnr > b.runnr)
    return true;
  else if (a.runnr < b.runnr)
    return false;

  // - Order by version (descending, natural string comparison).
  if (natcmpstr(b.version, a.version))
    return true;
  else if (natcmpstr(a.version, b.version))
    return false;

  // - Order by dataset (descending, natural string comparison).
  if (natcmpstr(a.dataset, b.dataset))
    return true;
  else if (natcmpstr(b.dataset, a.dataset))
    return false;

  // Should not come here, but give deterministic order.
  return &a < &b;
}

static void
sampleToJSON(const VisDQMSample &sample, std::string &result)
{
  result += "{'type':'";
  result += sampleTypeLabel[sample.type];
  result += "', 'run':'";
  result += StringFormat("%1").arg(sample.runnr);
  result += "', 'dataset':'";
  result += sample.dataset;
  result += "', 'version':'";
  result += sample.version;
  result += "'}";
}

static std::string
sampleToJSON(const VisDQMSample &sample)
{
  std::string result;
  result.reserve(64 + sample.dataset.size());
  sampleToJSON(sample, result);
  return result;
}

static VisDQMSample
sessionSample(const py::dict &session)
{
  VisDQMSample result;
  long type = py::extract<long>(session.get("dqm.sample.type"));
  result.runnr = py::extract<long>(session.get("dqm.sample.runnr"));
  result.dataset = py::extract<std::string>(session.get("dqm.sample.dataset"));

  if (type < SAMPLE_LIVE || type > SAMPLE_OFFLINE_MC)
    throw std::runtime_error(StringFormat
			     ("invalid session dqm.sample.type: %1")
			     .arg(type));
      
  result.type = (VisDQMSampleType) type;

  if (result.runnr < 0)
    throw std::runtime_error(StringFormat
			     ("invalid session dqm.sample.runnr: %1")
			     .arg(result.runnr));
  return result;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMSource
{
public:
  VisDQMSource(void)
    {}

  virtual ~VisDQMSource(void)
    {}

  virtual const char *
  plotter(void) const
    {
      return "unknown";
    }

  virtual void
  prepareSession(py::dict session)
    {}

  virtual void
  prescan(void)
    {}

  virtual void
  scan(VisDQMItems &result,
       const VisDQMSample &sample,
       VisDQMEventNum &current,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm,
       std::string *layoutroot,
       Regexp *rxlayout)
    {}

  virtual void
  samples(VisDQMSamples &samples)
    {}

private:
  VisDQMSource(VisDQMSource &);
  VisDQMSource &operator=(VisDQMSource &);
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMUnknownSource : public VisDQMSource
{
public:
  VisDQMUnknownSource(void)
    {}
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMLayoutSource : public VisDQMSource
{
  struct FileData
  {
    Time		mtime;
    VisDQMItems		items;
  };

  typedef std::map<Filename, FileData> FileDataMap;
  typedef std::list<Filename> FileNameList;

  pthread_mutex_t	lock_;
  FileDataMap		data_;
  FileNameList		files_;

public:
  VisDQMLayoutSource(py::list filenames)
    {
      pthread_mutex_init(&lock_, 0);
      py::stl_input_iterator<std::string> begin(filenames), end;
      files_.insert(files_.end(), begin, end);
    }

  ~VisDQMLayoutSource(void)
    {}

  virtual void
  prescan(void)
    {
      Lock gate(&lock_);
      for (FileNameList::iterator i = files_.begin(), e = files_.end(); i != e; ++i)
      {
	IOStatus st;
	if (! st.fill(*i))
	  data_.erase(*i);
	else
	{
	  FileData &d = data_[*i];
	  if (d.mtime != st.m_mtime)
	  {
	    d.mtime = st.m_mtime;
	    d.items.clear();
	    try
	    {
	      py::dict items;
	      py::dict locals;
	      locals["dqmitems"] = items;
	      py::exec_file(i->name(), py::import("GuiDQM").attr("__dict__"), locals);

	      py::stl_input_iterator<std::string> ki(items), ke;
	      for ( ; ki != ke; ++ki)
	      {
		shared_ptr<VisDQMItem> i(new VisDQMItem);
		i->flags = 0;
		i->version = st.m_mtime.ns();
		i->name = StringAtom(&stree, *ki);
		i->plotter = 0;
		buildParentNames(i->parents, i->name);
		d.items[i->name] = i;

		i->layout.reset(new VisDQMLayoutRows);
		py::object rows = py::extract<py::object>(items.get(*ki).attr("layout"));
		py::stl_input_iterator<py::object> ri(rows), re;
		for ( ; ri != re; ++ri)
		{
		  py::stl_input_iterator<py::object> ci(*ri), ce;
		  shared_ptr<VisDQMLayoutRow> row(new VisDQMLayoutRow);
		  for ( ; ci != ce; ++ci)
		  {
		    py::object pycol = *ci;
		    py::extract<std::string> pycolstr(pycol);
		    py::extract<py::dict> pycoldict(pycol);
		    shared_ptr<VisDQMLayoutItem> col(new VisDQMLayoutItem);
		    if (pycol.ptr() == Py_None)
		      // None, leave name as empty.
		      col->path = StringAtom(&stree, std::string());
		    else if (pycolstr.check())
		      // Basic string, use as a name.
		      col->path = StringAtom(&stree, pycolstr());
		    else if (pycoldict.check())
		    {
		      // Complete description, copy over.
		      py::dict colobj = pycoldict();
		      col->path = StringAtom(&stree, py::extract<std::string>(colobj.get("path")));
		      col->desc = py::extract<std::string>(colobj.get("description", ""));
		      if (colobj.has_key("draw"))
			translateDrawOptions(py::extract<py::dict>(colobj.get("draw")), col->drawopts);
		    }
		    else
		      logwarn()
			<< "layout element " << i->name.string()
			<< "[" << i->layout->size() << ", " << row->columns.size()
			<< "] is neither 'None', string nor a dict!\n";

		    row->columns.push_back(col);
		  }
		  
		  i->layout->push_back(row);
		}
	      }
	    }
	    catch (const py::error_already_set &e)
	    {
	      logerr() << "failed to update layouts from " << *i << std::endl;
	      PyErr_Print();
	    }
	  }
	}
      }
    }

  virtual void
  scan(VisDQMItems &result,
       const VisDQMSample &sample,
       VisDQMEventNum &current,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm,
       std::string *layoutroot,
       Regexp *rxlayout)
    {
      Lock gate(&lock_);
      FileDataMap::iterator fi, fe;
      VisDQMItems::iterator oi, oe;
      for (fi = data_.begin(), fe = data_.end(); fi != fe; ++fi)
      {
	FileData &fd = fi->second;
	result.resize(result.size() + fd.items.size());
	for (oi = fd.items.begin(), oe = fd.items.end(); oi != oe; ++oi)
	{
	  VisDQMItem &o = *oi->second;

	  if (layoutroot && rxlayout)
	  {
	    RegexpMatch m;
            if (rxlayout->match(o.name.string(), 0, 0, &m))
	    {
	      std::string root = m.matchString(o.name.string(), 1);
	      if (*layoutroot == root)
		;
	      else if (layoutroot->empty())
		*layoutroot = root;
	      else
		layoutroot->clear();
            }
          }

	  if (rxmatch && ! rxmatch->match(o.name.string()))
	    continue;

	  if (rxsearch && rxsearch->search(o.name.string()) < 0)
	    continue;

	  if (alarm && ((o.flags & VisDQMIndex::SUMMARY_PROP_REPORT_ALARM) != 0) != (*alarm == true))
	    continue;

	  result[o.name] = oi->second;
	}
      }
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
/** A background thread assisting #VisDQMLiveSource.  Listens to
    changes in live DQM data in upstream sources and maintains a
    up-to-date summary relevant for the web service. */
class VisDQMLiveThread : public DQMBasicNet
{
  static const uint64_t DEAD_PRUNE = 900000000000ULL; // 15 minutes
  VisDQMSource		*owner_;
  VisDQMItems		items_;
  Regexp		rxvar_;
  pthread_cond_t	reqrecv_;
  std::string		streamers_;

public:
  VisDQMLiveThread(VisDQMSource *owner, bool verbose, const std::string &host, int port)
    : DQMBasicNet("VisDQMLiveThread"),
      owner_(owner),
      rxvar_("<.*?>(i|f|s|qr)=(.*?)</.*?>")
    {
      buildStreamerInfo(streamers_);
      pthread_cond_init(&reqrecv_, 0);
      rxvar_.study();
      logme()
	<< "INFO: DQM live thread started"
	<< ", listening for data from "
	<< host << ":" << port << "\n";

      debug(verbose);
      delay(100);
      listenToCollector(host, port);
      start();
    }

  void
  stop(void)
    {
      shutdown();
    }

  StringPair
  fetch(const std::string &name)
    {
      // Build an internal object request.
      DataBlob req(3*sizeof(uint32_t) + name.size());
      uint32_t header[3] = { req.size(), DQM_MSG_GET_OBJECT, name.size() };
      memcpy(&req[0], header, sizeof(header));
      memcpy(&req[sizeof(header)], &name[0], name.size());

      // Build a fake request/reply structure with a phatom peer.
      Bucket reply;
      reply.next = 0;

      Peer phantom;
      phantom.peeraddr = "(internal request)";
      phantom.socket = 0;
      phantom.sendq = 0;
      phantom.sendpos = 0;
      phantom.mask = 0;
      phantom.source = false;
      phantom.update = false;
      phantom.updated = false;
      phantom.updatefull = false;
      phantom.updates = 0;
      phantom.waiting = 0;
      phantom.automatic = 0;

      // Issue the fake object request as if it was from a real peer,
      // and then wait for the response.  If we get the response right
      // here and now, queue it to the phantom responding peer.  Note
      // that we use the base class lock here for everything, including
      // as the condition variable mutex, to not miss notifications.
      lock();
      onMessage(&reply, &phantom, &req[0], req.size());
      if (! reply.data.empty())
      {
	Bucket **prev = &phantom.sendq;
	while (*prev)
	  prev = &(*prev)->next;

	*prev = new Bucket;
	(*prev)->next = 0;
	(*prev)->data.swap(reply.data);
      }

      // Wait for the response to arrive.  Note that we use the base
      // class lock as the mutex for the condition variable, so we
      // avoid missing notifications and will exit this loop with the
      // lock held in order for discard() and findObject() not begin
      // until releaseFromWait() and sendObjectToPeer() have completed.
      while (! phantom.sendq)
	  pthread_cond_wait(&reqrecv_, &lock_);

      // Got the object.  We don't need the reply packet since
      // we also know the object by now (if it exists).
      StringPair result;
      discard(phantom.sendq);
      if (Object *o = findObject(0, name))
      {
	if (o->flags & DQM_FLAG_SCALAR)
	{
	  DataBlob tmp;
	  uint32_t flags;
	  getScalarData(*o, tmp, result.second, flags);
	}
	else
	{
	  const char *begin = (const char *) &o->rawdata[0];
	  const char *end = begin + o->rawdata.size();
	  result.first = streamers_; // FIXME: from peer.
	  result.second = std::string(begin, end);
	}
      }
      unlock();

      return result;
    }

  void
  scan(VisDQMItems &result, VisDQMEventNum &current,
       Regexp *rxmatch, Regexp *rxsearch, bool *alarm)
    {
      lock();

      DQMNet::DataBlob data;
      VisDQMEventNumList eventnums;
      PeerMap::iterator pi, pe;
      ObjectMap::iterator oi, oe;
      VisDQMItems::iterator di, de;
      uint64_t now = Time::current().ns();

      // First mark all replica objects no longer known.
      for (di = items_.begin(), de = items_.end(); di != de; ++di)
	di->second->flags |= DQM_FLAG_ZOMBIE;

      // Scan still known objects, and update replica if necessary.
      for (pi = peers_.begin(), pe = peers_.end(); pi != pe; ++pi)
	for (oi = pi->second.objs.begin(), oe = pi->second.objs.end(); oi != oe; ++oi)
	{
	  Object &o = oi->second;
	  bool isnew = o.flags & DQM_FLAG_NEW;
	  StringAtom name(&stree, o.name);
	  shared_ptr<VisDQMItem> &i = items_[name];
	  if (! i)
	  {
	    i.reset(new VisDQMItem);
	    i->flags = VisDQMIndex::SUMMARY_PROP_TYPE_INVALID;
	    i->version = 0;
	    i->name = name;
	    i->plotter = owner_;
	    buildParentNames(i->parents, i->name);
	    isnew = true;
	  }

	  i->flags &= ~DQM_FLAG_ZOMBIE;
	  if (isnew || o.flags != i->flags)
	  {
	    o.flags &= ~DQM_FLAG_NEW;
	    i->flags &= ~(VisDQMIndex::SUMMARY_PROP_TYPE_MASK
			  | VisDQMIndex::SUMMARY_PROP_REPORT_MASK
			  | VisDQMIndex::SUMMARY_PROP_HAS_REFERENCE
			  | VisDQMIndex::SUMMARY_PROP_TAGGED);

	    if (o.flags & DQM_FLAG_REPORT_ERROR)
	      i->flags |= VisDQMIndex::SUMMARY_PROP_REPORT_ERROR;
	    if (o.flags & DQM_FLAG_REPORT_WARNING)
	      i->flags |= VisDQMIndex::SUMMARY_PROP_REPORT_WARN;
	    if (o.flags & DQM_FLAG_REPORT_OTHER)
	      i->flags |= VisDQMIndex::SUMMARY_PROP_REPORT_OTHER;

	    i->version = o.version;

	    if (o.flags & DQM_FLAG_SCALAR)
	      getScalarData(o, data, i->data, i->flags);
	  }
	}

      // Finally scan replicas again.  Remove too old dead objects as
      // we run into them.  Match remaining valid objects against the
      // caller's filters and add matching ones to the result set.
      result.resize(result.size() + items_.size());
      for (di = items_.begin(), de = items_.end(); di != de; )
      {
	VisDQMItem &i = *di->second;

	if (i.flags & DQM_FLAG_ZOMBIE)
	  i.flags |= DQM_FLAG_DEAD;

	if ((i.flags & DQM_FLAG_DEAD) && now - i.version > DEAD_PRUNE)
	{
	  items_.erase(di++);
	  continue;
	}

	if (isIntegerType(i.flags) && i.name.string().find("/EventInfo/i") != std::string::npos)
	  getEventInfoNum(i.name.string(), i.data.c_str(), eventnums);
	else if (isRealType(i.flags) && i.name.string().find("/EventInfo/ru") != std::string::npos)
	  getEventInfoNum(i.name.string(), i.data.c_str(), eventnums);

	if ((rxmatch && ! rxmatch->match(i.name.string()))
	    || (rxsearch && rxsearch->search(i.name.string()) < 0)
	    || (alarm && ((i.flags & VisDQMIndex::SUMMARY_PROP_REPORT_ALARM) != 0) != (*alarm == true)))
	{
	  ++di;
	  continue;
	}

	result[i.name] = di->second;
	++di;
      }

      // Pick greatest run/lumi/event number combo seen.
      setEventInfoNums(eventnums, current);

      unlock();
    }

protected:
  virtual void
  releaseFromWait(Bucket *msg, WaitObject &w, Object *o)
    {
      DQMBasicNet::releaseFromWait(msg, w, o);
      pthread_cond_broadcast(&reqrecv_);
    }

  virtual void
  sendObjectToPeer(Bucket *msg, Object &o, bool data, bool text)
    {
      DQMBasicNet::sendObjectToPeer(msg, o, data, text);
      pthread_cond_broadcast(&reqrecv_);
    }

  void
  getScalarData(Object &o, DataBlob &tmp, std::string &data, uint32_t &flags)
    {
      RegexpMatch m;
      tmp.clear();
      extractScalarData(tmp, o);
      tmp.push_back(0);
      if (rxvar_.match((const char *) &tmp[0], 0, 0, &m))
      {
	data = m.matchString((const char *) &tmp[0], 2);
	switch (tmp[m.matchPos(1)])
	{
	case 'i': flags |= VisDQMIndex::SUMMARY_PROP_TYPE_INT; break;
	case 'f': flags |= VisDQMIndex::SUMMARY_PROP_TYPE_REAL; break;
	case 's': flags |= VisDQMIndex::SUMMARY_PROP_TYPE_STRING; break;
	case 'q': ASSERT(false); break;
	}
      }
    }
};

class VisDQMLiveSource : public VisDQMSource
{
  std::string mydataset_;
  VisDQMLiveThread *thread_;

public:
  VisDQMLiveSource(py::object gui, py::dict opts)
    : mydataset_(py::extract<std::string>(opts["dataset"])),
      thread_(new VisDQMLiveThread(this,
				   py::extract<bool>(opts["verbose"]),
				   py::extract<std::string>(opts["dqmhost"]),
				   py::extract<int>(opts["dqmport"])))
    {}

  ~VisDQMLiveSource(void)
    {
      if (thread_)
	stop();
    }

  virtual const char *
  plotter(void) const
    {
      return "live";
    }

  StringPair
  data(const std::string &name)
    {
      PyReleaseInterpreterLock nogil;
      return thread_->fetch(name);
    }
  
  void
  stop(void)
    {
      thread_->stop();
      delete thread_;
      thread_ = 0;
    }

  virtual void
  prepareSession(py::dict session)
    {
      if (! session.has_key("dqm.sample.type"))
      {
	session["dqm.sample.runnr"] = 0L;
	session["dqm.sample.dataset"] = mydataset_;
	session["dqm.sample.type"] = long(SAMPLE_LIVE);
      }
    }

  virtual void
  scan(VisDQMItems &result,
       const VisDQMSample &sample,
       VisDQMEventNum &current,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm,
       std::string *,
       Regexp *)
    {
      if (sample.type == SAMPLE_LIVE)
	thread_->scan(result, current, rxmatch, rxsearch, alarm);
    }

  virtual void
  samples(VisDQMSamples &samples)
    {
      samples.push_back(VisDQMSample());
      VisDQMSample &s = samples.back();
      s.type = SAMPLE_LIVE;
      s.runnr = 0;
      s.dataset = mydataset_;
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
/** A background thread assisting #VisDQMArchiveSource.  Prompts
    the source to revalidate its index data every so often. */
class VisDQMArchiveWatch
{
  pthread_mutex_t	lock_;
  pthread_t		thread_;
  Filename		path_;
  bool			stop_;
  bool			changed_;
  Time			mtime_;

public:
  VisDQMArchiveWatch(const Filename &path)
    : thread_((pthread_t) -1),
      path_(path),
      stop_(false),
      changed_(false),
      mtime_(0)
    {
      pthread_mutex_init(&lock_, 0);
      checkForUpdate();
      pthread_create(&thread_, 0, &run, this);
    }

  bool
  running(void)
    {
      return thread_ != (pthread_t) -1;
    }

  void
  stop(void)
    {
      pthread_mutex_lock(&lock_);
      stop_ = true;
      pthread_mutex_unlock(&lock_);
      pthread_join(thread_, 0);
      thread_ = (pthread_t) -1;
    }

  bool
  changed(void)
    {
      pthread_mutex_lock(&lock_);
      bool changed = changed_;
      pthread_mutex_unlock(&lock_);
      return changed;
    }

  void
  clear(void)
    {
      pthread_mutex_lock(&lock_);
      changed_ = false;
      pthread_mutex_unlock(&lock_);
    }

private:
  static void *
  run(void *arg)
    { return ((VisDQMArchiveWatch *)arg)->dorun(); }

  void *
  dorun(void)
    {
      bool stop = false;
      do
      {
	for (int i = 0; i < 300 && ! stop; ++i)
	{
	  TimeInfo::usleep(100);

	  pthread_mutex_lock(&lock_);
	  stop = stop_;
	  pthread_mutex_unlock(&lock_);
	}
	
	pthread_mutex_lock(&lock_);
	checkForUpdate();
	pthread_mutex_unlock(&lock_);
      } while (! stop);

      return 0;
    }

  void
  checkForUpdate(void)
    {
      IOStatus st;
      Filename gen(path_, "generation");
      if (st.fill(gen.name()) && st.m_mtime != mtime_)
      {
	changed_ = true;
	mtime_ = st.m_mtime;
      }
    }
};

class VisDQMArchiveSource : public VisDQMSource
{
  typedef shared_ptr<VisDQMFile>		VisDQMFilePtr;
  typedef std::vector<size_t>			IdMap;
  typedef std::vector<VisDQMIndex::Sample>	SampleList;
  typedef std::map<uint32_t, VisDQMFilePtr>	FileMap;

  Regexp		rxonline_;
  pthread_rwlock_t	lock_;
  bool			retry_;
  Filename		path_;
  VisDQMArchiveWatch	watch_;
  VisDQMCache		cache_;
  VisDQMIndex		index_;
  SampleList		samples_;
  StringAtomTree	vnames_; // (10000);
  StringAtomTree	dsnames_; // (10000);
  StringAtomTree	objnames_; // (400000);
  StringAtomTree	streamers_; // (100);
  FileMap		infoFiles_;
  FileMap		dataFiles_;

  // Helper utility to close a DQM file before destroying it.
  static void
  closeDQMFile(VisDQMFile *f)
    {
      if (f && f->path())
	f->close();
    }

  // Utility function to reset current data contents.  Must be called
  // with write lock held.  Does not terminate transaction.
  void
  reset(void)
    {
      dataFiles_.clear();
      infoFiles_.clear();
      streamers_.clear();
      objnames_.clear();
      dsnames_.clear();
      vnames_.clear();
      samples_.clear();
      watch_.clear();
      retry_ = false;
    }

  // Load the master catalogue from an index.
  void
  loadIndex(void)
    {
      // FIXME: Try doing the index re-read in the watcher thread
      // and simply swap 'current' values in place here, so we do
      // not become unnecessarily blocked waiting for the I/O while
      // serving HTTP requests.  We do have to block when file
      // open errors have occurred however.
      VisDQMFile *master = 0;
      index_.beginRead(master);
      for (VisDQMFile::ReadHead rdhead(master, 0); ! rdhead.isdone(); rdhead.next())
      {
	void *begin;
	void *end;
	uint64_t key;
	uint64_t hipart;
	uint64_t lopart;

	rdhead.get(&key, &begin, &end);
	hipart = key & 0xffffffff00000000ull;
	lopart = key & 0x00000000ffffffffull;
	if (hipart == VisDQMIndex::MASTER_SAMPLE_RECORD)
	  samples_.push_back(*(const VisDQMIndex::Sample *)begin);
	else
	  break;
      }

      readStrings(dsnames_, master, VisDQMIndex::MASTER_DATASET_NAME);
      readStrings(vnames_, master, VisDQMIndex::MASTER_CMSSW_VERSION);
      readStrings(objnames_, master, VisDQMIndex::MASTER_OBJECT_NAME);
      readStrings(streamers_, master, VisDQMIndex::MASTER_TSTREAMERINFO);
      index_.finishRead();
    }

  // Report a file read failure.  Returns true if reading should
  // be abandoned: more than 10 tries have failed to open a file.
  bool
  fileReadFailure(int ntries, const char *message)
    {
      bool abandon = (ntries < 0 || ntries > 10);
      loginfo()
	<< (ntries >= 0 ? ntries : 1)
	<< " transient error" << (ntries > 1 ? "s" : "")
	<< " while reading data, " << (abandon ? "giving up" : "retrying")
	<< "; error was: " << message << std::endl;
      pthread_rwlock_wrlock(&lock_);
      retry_ = true;
      pthread_rwlock_unlock(&lock_);
      return abandon;
    }
  
  // Prepare to read from the index. Reloads the index if there have
  // been past file read errors, or if the index has changed.
  void
  maybeReloadIndex(void)
    {
      // First determine if we should retry index open.
      pthread_rwlock_rdlock(&lock_);
      bool retry = retry_;
      pthread_rwlock_unlock(&lock_);

      // If we've seen file read errors, force index refresh.  But
      // avoid storming-heard several thread reload by checking
      // once again after we obtain the exclusive write lock.
      if (retry)
      {
	WRLock wrgate(&lock_);
	if (retry_)
	{
	  reset();
	  loadIndex();
	}
      }

      // If the index changed we should reload, so do so now.
      // Again avoid storming herd repeated updates.  Note that if
      // the previous block updated index, it cleared the watcher.
      if (watch_.changed())
      {
	WRLock wrgate(&lock_);
	if (watch_.changed())
	{
	  reset();
	  loadIndex();
	}
      }
    }

  // Open a possibly cached data file.  This must be called with a
  // read lock on "this" held; it will upgrade the lock to a write
  // lock if that is necessary to open the file.
  VisDQMFilePtr
  open(uint16_t kind, const uint32_t *file)
    {
      FileMap &files = (kind == VisDQMIndex::MASTER_FILE_INFO
			? infoFiles_ : dataFiles_);
      FileMap::iterator pos = files.find(file[kind]);
      if (pos == files.end())
      {
	pthread_rwlock_unlock(&lock_);
	pthread_rwlock_wrlock(&lock_);
	pos = files.find(file[kind]);
      }

      if (pos == files.end())
      {
	VisDQMFilePtr ptr(index_.open(kind,
				      file[kind] >> 16,
				      file[kind] & 0xffff,
				      VisDQMFile::OPEN_READ),
			  &closeDQMFile);
	files.insert(FileMap::value_type(file[kind], ptr));
	return ptr;
      }
      else
	return pos->second;
    }

  SampleList::iterator
  findSample(const VisDQMSample &sample)
    {
      SampleList::iterator i;
      SampleList::iterator e;
      for (i = samples_.begin(), e = samples_.end(); i != e; ++i)
      {
	const std::string &dsname = dsnames_.key(i->datasetNameIdx+1);
	if (false
	    || (sample.type <= SAMPLE_OFFLINE_DATA
		&& sample.runnr == i->runNumber
		&& sample.dataset == dsname)
	    || (sample.type == SAMPLE_OFFLINE_RELVAL
		&& i->runNumber < 0
		&& sample.runnr == 0
		&& sample.dataset == dsname)
	    || (sample.type == SAMPLE_OFFLINE_MC
		&& i->runNumber == 0
		&& sample.runnr == 0
		&& sample.dataset == dsname))
	  break;
      }

      return i;
    }

public:
  VisDQMArchiveSource(py::object gui, py::dict opts)
    : rxonline_(py::extract<std::string>(opts["rxonline"])),
      retry_(false),
      path_(Filename(py::extract<std::string>(opts["index"]))),
      watch_(path_),
      index_(path_, &cache_),
      vnames_(10000),
      dsnames_(10000),
      objnames_(1000000),
      streamers_(100)
    {
      rxonline_.study();
      pthread_rwlock_init(&lock_, 0);
    }

  ~VisDQMArchiveSource(void)
    {
      if (watch_.running())
	watch_.stop();
    }

  virtual const char *
  plotter(void) const
    {
      return "archive";
    }

  void
  stop(void)
    {
      if (watch_.running())
	watch_.stop();
    }

  /** Return streamer and data blob to send to render back-end. */
  StringPair
  data(VisDQMObjType objtype,
       VisDQMSampleType sampletype,
       long runnr,
       const std::string &dataset,
       const std::string &name)
    {
      PyReleaseInterpreterLock nogil;
      VisDQMSample sample = { sampletype, runnr, dataset };
      StringPair result;

      // Keep retrying until we can successfully read the data.
      for (int ntries = 1; true; ++ntries)
      {
	try
	{
	  // Reload index if necessary.
	  maybeReloadIndex();

	  // Locate the requested sample.
	  RDLock rdgate(&lock_);
	  SampleList::const_iterator si = findSample(sample);
	  uint32_t objindex = StringAtom(&objnames_, name, StringAtom::TestOnly).index();

	  // Check if we found what we are looking for.  If yes, try
	  // opening the data file and return the streamed data for
	  // the requested object; if this fails we'll end up in a
	  // catch statement which will loop around.
	  if (si != samples_.end() && objindex != 0)
	  {
	    // FIXME: handle keys other than run summary?
	    uint64_t keyidx = (uint64_t(si - samples_.begin()) << 44) | (objindex-1);
	    if (objtype == OBJ_SCALAR)
	    {
	      VisDQMFilePtr file(open(VisDQMIndex::MASTER_FILE_INFO, si->files));
	      VisDQMFile::ReadHead rdhead(file.get(), keyidx);
	      if (! rdhead.isdone())
	      {
		uint64_t key;
		void *begin;
		void *end;
		rdhead.get(&key, &begin, &end);
		if (key == keyidx)
		{
		  VisDQMIndex::Summary *s = (VisDQMIndex::Summary *)begin;
		  result.second.append((const char *) (s+1), s->dataLength-1);
		  ASSERT(isScalarType(s->properties));
		}
	      }
	    }
	    else if (objtype == OBJ_ROOTOBJ)
	    {
	      VisDQMFilePtr file(open(VisDQMIndex::MASTER_FILE_DATA, si->files));
	      VisDQMFile::ReadHead rdhead(file.get(), keyidx);
	      if (! rdhead.isdone())
	      {
		uint64_t key;
		void *begin;
		void *end;
		rdhead.get(&key, &begin, &end);
		if (key == keyidx)
		{
		  result.first = streamers_.key(si->streamerInfoIdx+1);
		  result.second.append((const char *) begin, (const char *) end);
		}
	      }
	    }
	  }
	  
	  // Not found, return null data.
	  break;
	}
	catch (Error &e)
	{
	  if (fileReadFailure(ntries, e.explain().c_str()))
	    break;
	}
	catch (std::exception &e)
	{
	  if (fileReadFailure(ntries, e.what()))
	    break;
	}
	catch (...)
	{
	  if (fileReadFailure(ntries, "(unknown error)"))
	    break;
	}
      }

      return result;
    }

  // Refresh data from the database for a run.  Checks if the master
  // database exists, and if so, if it has changed relative to the
  // last copy and the time we last read in the data.  If so, discards
  // any existing cached data and copies the file again.  Then checks
  // to see if the data for the requested run is already cached, and
  // if not, reads it in.  Also discards cached data for runs that
  // have not been accessed recently, where "recent" is MAX_CACHE_LIFE
  // seconds.
  void
  update(VisDQMItems &items,
	 const VisDQMSample &sample,
	 VisDQMEventNum &current,
	 Regexp *rxmatch,
	 Regexp *rxsearch,
	 bool *alarm)
    {
      // No point in even trying unless this is archived data.
      if (sample.type < SAMPLE_ONLINE_DATA)
	return;
      
      // Keep retrying until we can successfully read the data.
      for (int ntries = 1; true; ++ntries)
      {
	try
	{
	  // Reload index if necessary.
	  maybeReloadIndex();

	  // Locate the requested sample.
	  RDLock rdgate(&lock_);
	  SampleList::const_iterator si = findSample(sample);

	  // Return with empty "items" if we didn't find the sample.
	  if (si == samples_.end())
	    return;

	  // Found the sample, try opening the data file.  If this
	  // fails we will end up in the catch statement which will
	  // loop around to retry a certain number of times.  The
	  // stack unwinding automatically handles lock release.
	  VisDQMFilePtr file(open(VisDQMIndex::MASTER_FILE_INFO, si->files));

	  // Read the summary for the requested sample.
	  VisDQMEventNumList eventnums;
	  uint64_t keyidx = (si - samples_.begin());
	  for (VisDQMFile::ReadHead rdhead(file.get(), keyidx << 44);
	       ! rdhead.isdone(); rdhead.next())
	  {
	    uint64_t key;
	    void *begin;
	    void *end;
	    rdhead.get(&key, &begin, &end);
	    // FIXME: handle keys other than run summary?
	    uint64_t keyparts[4] = { (key >> 44) & 0xfffff, (key >> 40) & 0xf,
				     (key >> 20) & 0xfffff, key & 0xfffff };
	    if (keyparts[0] != keyidx || keyparts[1] != 0)
	      break;

	    VisDQMIndex::Summary *s = (VisDQMIndex::Summary *) begin;
	    uint32_t report = s->properties & VisDQMIndex::SUMMARY_PROP_REPORT_MASK;
	    const char *data = (s->dataLength ? (const char *) (s+1) : 0);
	    const char *qdata = (s->qtestLength ? (const char *) (s+1) + s->dataLength : 0);
	    const std::string &name = objnames_.key(keyparts[3]+1);

	    // Update run number information.
	    if (isIntegerType(s->properties) && name.find("/EventInfo/i") != std::string::npos)
	      getEventInfoNum(name, data, eventnums);
	    else if (isRealType(s->properties) && name.find("/EventInfo/ru") != std::string::npos)
	      getEventInfoNum(name, data, eventnums);

	    // If it doesn't match search criteria, skip it.
	    if ((rxmatch && ! rxmatch->match(name))
		|| (rxsearch && rxsearch->search(name) < 0)
		|| (alarm && (report != 0) != (*alarm == true)))
	      continue;

	    // We want to keep it, construct an item for use.
	    shared_ptr<VisDQMItem> i(new VisDQMItem);
	    i->flags = s->properties;
	    i->version = si->lastImportTime;
	    i->name = StringAtom(&stree, name);
	    i->plotter = this;
	    buildParentNames(i->parents, i->name);
	    if (s->dataLength)
	      i->data = data;

	    items[i->name] = i;
	  }

	  // Pick greatest run/lumi/event number combo seen.
	  setEventInfoNums(eventnums, current);

	  // Successfully read the data, return.
	  return;
	}
	catch (Error &e)
	{
	  if (fileReadFailure(ntries, e.explain().c_str()))
	    return;
	}
	catch (std::exception &e)
	{
	  if (fileReadFailure(ntries, e.what()))
	    return;
	}
	catch (...)
	{
	  if (fileReadFailure(ntries, "(unknown error)"))
	    return;
	}
      }
    }

  // Initialise a new session.  Select the most recent sample.
  virtual void
  prepareSession(py::dict session)
    {
      if (! session.has_key("dqm.sample.type"))
      {
	// Make sure we are up to date.
	VisDQMItems items;
	VisDQMSample sample = { SAMPLE_ONLINE_DATA, 0 };
	VisDQMEventNum current = { "", -1, -1, -1, -1 };
	update(items, sample, current, 0, 0, 0);

	// Scan samples for the most recent one.
	RDLock rdgate(&lock_);
	SampleList::const_iterator newest;
	SampleList::const_iterator si;
	SampleList::const_iterator se;
	for (newest = si = samples_.begin(), se = samples_.end(); si != se; ++si)
	  if (si->runNumber > newest->runNumber)
	    newest = si;
	  else if (si->runNumber == newest->runNumber
		   && si->lastImportTime > newest->lastImportTime)
	    newest = si;

	if (newest != se)
	{
	  VisDQMSampleType type
	    = (newest->runNumber < 0 ? SAMPLE_OFFLINE_RELVAL
	       : newest->runNumber == 0 ? SAMPLE_OFFLINE_MC
	       : rxonline_.search(dsnames_.key(newest->datasetNameIdx+1)) < 0
	       ? SAMPLE_OFFLINE_DATA : SAMPLE_ONLINE_DATA);

	  session["dqm.sample.runnr"] = (newest->runNumber > 0 ? long(newest->runNumber) : 0L);
	  session["dqm.sample.dataset"] = dsnames_.key(newest->datasetNameIdx+1);
	  session["dqm.sample.type"] = long(type);
	}
      }
    }

  // Extract data from this source.  Work is done by update().
  virtual void
  scan(VisDQMItems &result,
       const VisDQMSample &sample,
       VisDQMEventNum &current,
       Regexp *rxmatch,
       Regexp *rxsearch,
       bool *alarm,
       std::string *,
       Regexp *)
    {
      update(result, sample, current, rxmatch, rxsearch, alarm);
    }

  virtual void
  samples(VisDQMSamples &samples)
    {
      try
      {
	// Reload index if necessary.
	maybeReloadIndex();

	// Locate the requested sample.
	RDLock rdgate(&lock_);
	SampleList::const_iterator si;
	SampleList::const_iterator se;
	samples.reserve(samples.size() + samples_.size());
	for (si = samples_.begin(), se = samples_.end(); si != se; ++si)
	{
	  samples.push_back(VisDQMSample());
	  VisDQMSample &s = samples.back();
	  s.dataset = dsnames_.key(si->datasetNameIdx+1);
	  if (si->runNumber < 0)
	    s.version = vnames_.key(-si->runNumber);
	  s.runnr = (si->runNumber >= 0 ? si->runNumber : 0);
	  s.type = (si->runNumber < 0 ? SAMPLE_OFFLINE_RELVAL
		    : si->runNumber == 0 ? SAMPLE_OFFLINE_MC
		    : rxonline_.search(s.dataset) < 0 ? SAMPLE_OFFLINE_DATA
		    : SAMPLE_ONLINE_DATA);
	}
      }
      catch (Error &e)
      {
	fileReadFailure(-1, e.explain().c_str());
      }
      catch (std::exception &e)
      {
	fileReadFailure(-1, e.what());
      }
      catch (...)
      {
	fileReadFailure(-1, "(unknown error)");
      }
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMWorkspace
{
protected:
  py::object	gui_;
  double	guiTimeStamp_;
  std::string	guiServiceName_;
  std::string	guiServiceListJSON_;
  std::string	guiWorkspaceListJSON_;
  std::string	workspaceName_;
  void		(*snapdump_)(const char *);

public:
  VisDQMWorkspace(py::object gui, const std::string &name)
    : gui_(gui),
      guiTimeStamp_(py::extract<double>(gui.attr("stamp"))),
      guiServiceName_(py::extract<std::string>(gui.attr("serviceName"))),
      workspaceName_(name),
      snapdump_(0)
    {
      if (void *sym = dlsym(0, "igprof_dump_now"))
        snapdump_ = (void(*)(const char *)) sym;
    }

  virtual ~VisDQMWorkspace(void)
    {}

  virtual std::string
  state(py::dict session)
    {
      return "{}";
    }

  virtual void
  profilesnap(void)
    {
      if (snapdump_)
      {
        char tofile[256];
        struct timeval tv;
        gettimeofday(&tv, 0);
        if (snprintf(tofile, sizeof(tofile), "|gzip -c>igprof.dqmgui.%ld.%f.gz", 
	             (long) getpid(), tv.tv_sec + 1e-6*tv.tv_usec) < int(sizeof(tofile)))
	  snapdump_(tofile);
      }
    }

protected:
  template <class T> T
  workspaceParam(const py::dict &session, const char *key)
    {
      py::dict d = py::extract<py::dict>(session.get(key));
      return py::extract<T>(d.get(this->workspaceName_));
    }
      
  template <class T> T
  workspaceParam(const py::dict &session, const char *key, const T &def)
    {
      py::dict d = py::extract<py::dict>(session.get(key));
      return py::extract<T>(d.get(this->workspaceName_, def));
    }

  template <class T> T
  workspaceParamOther(const py::dict &session, const char *key, const std::string &wspace)
    {
      py::dict d = py::extract<py::dict>(session.get(key));
      return py::extract<T>(d.get(wspace));
    }
      
  template <class T> T
  workspaceParamOther(const py::dict &session, const char *key, const T &def, const std::string &wspace)
    {
      py::dict d = py::extract<py::dict>(session.get(key));
      return py::extract<T>(d.get(wspace, def));
    }

  // Produce a standard response.
  std::string
  makeResponse(const std::string &state,
	       int interval,
	       VisDQMEventNum &current,
	       const std::string &services,
	       const std::string &workspace,
	       const std::string &workspaces,
	       const std::string &submenu,
	       const VisDQMSample &sample,
	       const std::string &filter,
	       const std::string &reference,
	       const std::string &rxstr,
	       const std::string &rxerr,
	       size_t rxmatches,
	       Time startTime)
  {
    StringFormat result
      = StringFormat("([{'kind':'AutoUpdate', 'interval':%1, 'stamp':%2, 'serverTime':%17},"
		     "{'kind':'DQMHeaderRow', 'run':\"%3\", 'lumi':\"%4\", 'event':\"%5\","
		     " 'runstart':\"%6\", 'service':'%7', 'services':[%8], 'workspace':'%9',"
		     " 'workspaces':[%10], 'view':{'show':'%11','sample': %12, 'filter':'%13',"
		     " 'reference':'%14', 'search':%15}},%16])")
	.arg(interval)
	.arg(guiTimeStamp_, 0, 'f')
	.arg(current.runnr < 0 ? std::string("(None)")
	     : current.runnr <= 1 ? std::string("(Simulated)")
	     : thousands(StringFormat("%1").arg(current.runnr)))
	.arg(current.luminr < 0 ? std::string("(None)")
	     : thousands(StringFormat("%1").arg(current.luminr)))
	.arg(current.eventnr < 0 ? std::string("(None)")
	     : thousands(StringFormat("%1").arg(current.eventnr)))
	.arg(formatStartTime(current.runstart))
	.arg(guiServiceName_)
	.arg(services)
	.arg(workspace)
	.arg(workspaces)
	.arg(submenu)
	.arg(sampleToJSON(sample))
	.arg(filter)
	.arg(reference)
	.arg(StringFormat("{'pattern':%1, 'valid':%2, 'nmatches':%3, 'error':%4}")
	     .arg(stringToJSON(rxstr))
	     .arg(rxerr.empty() ? 1 : 0)
	     .arg(rxmatches)
	     .arg(stringToJSON(rxerr)))
	.arg(state);

    return result.arg((Time::current() - startTime).ns() * 1e-6, 0, 'f');
  }

  // Get the list of native VisDQMSources configured in the GUI.
  void
  sources(std::vector<VisDQMSource *> &into)
    {
      py::list sources = py::extract<py::list>(gui_.attr("sources"));
      py::stl_input_iterator<py::object> i(sources), e;
      for ( ; i != e; ++i)
      {
	py::extract<VisDQMSource *> x(*i);
	if (x.check())
	  into.push_back(x());
      }
    }

  // Build and return a cached JSON representation of the DQM GUI
  // service list.  This must be called when it's safe to access
  // python.  Will build the list just once.
  const std::string &
  serviceListJSON(void)
    {
      if (guiServiceListJSON_.empty())
      {
	StringList parts;
	parts.reserve(100);

	py::list services = py::extract<py::list>(gui_.attr("services"));
	for (py::stl_input_iterator<py::tuple> i(services), e; i != e; ++i)
	  parts.push_back(StringFormat("{'title':'%1', 'href':'%2'}")
			  .arg(py::extract<std::string>((*i)[0]))
			  .arg(py::extract<std::string>((*i)[1])));
	guiServiceListJSON_ = StringOps::join(parts, ", ");
      }

      return guiServiceListJSON_;
    }

  // Build and return a cached JSON representation of the DQM GUI
  // workspace object list.  This must be called when it's safe to
  // access python.  Will build the list just once.
  const std::string &
  workspaceListJSON(void)
    {
      if (guiWorkspaceListJSON_.empty())
      {
	StringList parts;
	parts.reserve(100);
	py::list workspaces = py::extract<py::list>(gui_.attr("workspaces"));
	for (py::stl_input_iterator<py::object> i(workspaces), e; i != e; ++i)
	{
	  int         rank(py::extract<int>(i->attr("rank")));
	  std::string name(py::extract<std::string>(i->attr("name")));
	  std::string category(py::extract<std::string>(i->attr("category")));
	  std::string lower(name);
	  std::transform(lower.begin(), lower.end(), lower.begin(), tolower);
	  parts.push_back(StringFormat("{'title':'%1', 'label':'%2', 'category':'%3', 'rank':%4}")
			  .arg(name).arg(lower).arg(category).arg(rank));
	}
	guiWorkspaceListJSON_ = StringOps::join(parts, ", ");
      }

      return guiWorkspaceListJSON_;
    }

  // Get the contents currently shown in this session.
  void
  buildContents(const std::vector<VisDQMSource *> &sources,
		VisDQMItems &contents,

		StringAtomSet &myobjs,
		const VisDQMSample &sample,
		VisDQMEventNum &current,

		Regexp *rxmatch,
		Regexp *rxsearch,
		bool *alarm,

		std::string *layoutroot,
		Regexp *rxlayout)
    {
      // Now get filtered contents from all the sources.
      for (size_t i = 0, e = sources.size(); i != e; ++i)
	sources[i]->scan(contents, sample, current,
			 rxmatch, rxsearch, alarm,
			 layoutroot, rxlayout);
      

      // Add in the quick collection, filtered against rx if there is
      // one, but not the general workspace filter.
      StringAtomSet::iterator mi, me;
      for (mi = myobjs.begin(), me = myobjs.end(); mi != me; ++mi)
      {
	std::string trailer(mi->string(), StringOps::rfind(mi->string(), '/')+1);
	StringAtom label(&stree, "Quick collection/" + trailer);
	if (! rxsearch || rxsearch->search(label.string()) >= 0)
	{
	  shared_ptr<VisDQMItem> i(new VisDQMItem);
	  i->flags = 0;
	  i->version = 0;
	  i->name = label;
	  i->plotter = 0;
	  buildParentNames(i->parents, i->name);

	  VisDQMItems::iterator pos = contents.find(*mi);
	  if (pos != contents.end() && pos->second->layout)
	    i->layout = pos->second->layout;
	  else
	  {
	    shared_ptr<VisDQMLayoutRows> layout(new VisDQMLayoutRows);
	    shared_ptr<VisDQMLayoutRow> row(new VisDQMLayoutRow);
	    shared_ptr<VisDQMLayoutItem> column(new VisDQMLayoutItem);
	    column->path = *mi;
	    layout->push_back(row);
	    row->columns.push_back(column);
	    i->layout = layout;
	  }

	  contents[label] = i;
	}
      }
    }

  // Utility method to build a look-up table of shown objects.
  void
  buildShown(const VisDQMItems &contents,
	     StringAtomSet &shown,
	     const std::string &root)
    {
      std::string myroot;
      if (! root.empty())
      {
	myroot.reserve(root.size() + 2);
	myroot += root;
	myroot += '/';
      }

      VisDQMItems::const_iterator ci, ce;
      for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
	if (ci->first.string().size() > myroot.size()
	    && ci->first.string().compare(0, myroot.size(), myroot) == 0
	    && ci->first.string().find('/', myroot.size()) == std::string::npos)
	  shown.insert(ci->first);
    }

  // Update the object metrics for the items in @a contents.
  // "Shown" means the object is included in canvas listing.
  void
  updateStatus(const VisDQMItems &contents,
	       VisDQMStatusMap &status)
    {
      VisDQMStatus empty;
      empty.nleaves = 0;
      empty.nalarm = 0;
      empty.nlive = 0;

      VisDQMStatus leaf;
      leaf.nleaves = 1;
      leaf.nalarm = 0;
      leaf.nlive = 0;

      status[StringAtom(&stree, "")] = empty;
      status[StringAtom(&stree, "Quick collection")] = empty;

      status.resize(10*contents.size());
      VisDQMItems::const_iterator ci, ce, lpos;
      for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
      {
	// Initialise object data.
	VisDQMItem &o = *ci->second;
	VisDQMStatus &s = status[ci->first];
	s.nleaves = 1;
	s.nalarm = ((o.flags & VisDQMIndex::SUMMARY_PROP_REPORT_ALARM) ? 1 : 0);
	s.nlive = ((o.flags & DQMNet::DQM_FLAG_DEAD) ? 0 : 1);

	// If this is a layout, resolve it.
	if (o.layout)
	{
	  VisDQMLayoutRows &rows = *o.layout;
	  for (size_t nrow = 0; nrow < rows.size(); ++nrow)
	  {
	    VisDQMLayoutRow &row = *rows[nrow];
	    for (size_t ncol = 0; ncol < row.columns.size(); ++ncol)
	    {
	      VisDQMLayoutItem &col = *row.columns[ncol];
	      if ((lpos = contents.find(col.path)) != contents.end())
		s.nalarm += ((lpos->second->flags & VisDQMIndex::SUMMARY_PROP_REPORT_ALARM) ? 1 : 0);
	      else if (! col.path.string().empty())
		s.nlive = 0;
	    }
	  }
	}

	// Update data to "virtual parents."
	for (size_t i = 0, e = o.parents.size(); i != e; ++i)
	{
	  VisDQMStatusMap::iterator pos = status.find(o.parents[i]);
	  if (pos == status.end())
	    pos = status.insert(VisDQMStatusMap::value_type(o.parents[i], empty)).first;
	  pos->second.nleaves += s.nleaves;
	  pos->second.nalarm += s.nalarm;
	  pos->second.nlive += s.nlive;
	}
      }
    }

private:
  VisDQMWorkspace(VisDQMWorkspace &);
  VisDQMWorkspace &operator=(VisDQMWorkspace &);
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMSummaryWorkspace : public VisDQMWorkspace
{
  typedef std::map<std::string, std::string> KeyValueMap;
  typedef std::map<std::string, KeyValueMap> SummaryMap;

  Regexp rxevi_;
public:
  VisDQMSummaryWorkspace(py::object gui, const std::string &name)
    : VisDQMWorkspace(gui, name),
      rxevi_("(.*?)/EventInfo/(.*)")
    {
      rxevi_.study();
    }

  virtual ~VisDQMSummaryWorkspace(void)
    {}

  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      const std::string &services = serviceListJSON();
      const std::string &workspaces = workspaceListJSON();

      VisDQMSample sample(sessionSample(session));
      std::string qplot(py::extract<std::string>(session.get("dqm.qplot", "")));
      std::string filter(py::extract<std::string>(session.get("dqm.filter")));
      std::string reference(py::extract<std::string>(session.get("dqm.reference")));
      std::string submenu(py::extract<std::string>(session.get("dqm.submenu")));
      std::string rxstr(py::extract<std::string>(session.get("dqm.search")));
      std::string rxerr;
      shared_ptr<Regexp> rxsearch;
      makerx(rxstr, rxsearch, rxerr, Regexp::IgnoreCase);

      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	PyReleaseInterpreterLock nogil;

	VisDQMItems		contents;
	StringAtomSet		myobjs; // deliberately empty, not needed
	VisDQMItems::iterator	ci, ce;
	VisDQMStatusMap		status;
	RegexpMatch		m;
	std::string		plotter;
	SummaryMap		summary;
	VisDQMEventNum		current = { "", -1, -1, -1, -1 };
	bool			alarms = false;
	bool			*qalarm = 0;

	if (filter == "all")
	  qalarm = 0;
	else if (filter == "alarms")
	  alarms = true, qalarm = &alarms;
	else if (filter == "nonalarms")
	  alarms = false, qalarm = &alarms;

	buildContents(srclist, contents, myobjs, sample,
		      current, 0, rxsearch.get(), 0, 0, 0);
	updateStatus(contents, status);

	for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
	{
	  VisDQMItem &o = *ci->second;
	  if (! isScalarType(o.flags))
	    continue;

	  m.reset();
	  if (! rxevi_.match(o.name.string(), 0, 0, &m))
	    continue;

	  std::string system(m.matchString(o.name.string(), 1));
	  std::string param(m.matchString(o.name.string(), 2));

	  if (rxsearch && rxsearch->search(system) < 0)
	    continue;

	  // Alarm filter. Somewhat convoluted as we need to dig out the
	  // report summary value, and use the value known by JavaScript
	  // to check for items passing an alarm filter here.
	  if (qalarm)
	  {
	    StringAtom reportSummary(&stree,
				     system + "/EventInfo/reportSummary",
				     StringAtom::TestOnly);

	    VisDQMItems::iterator pos = contents.find(reportSummary);
	    if (pos != ce && isRealType(pos->second->flags))
	    {
	      double value = atof(pos->second->data.c_str());
	      if ((value >= 0 && value < 0.95) != alarms)
		continue;
	    }
	  }

	  KeyValueMap &info = summary[system];
	  if (info.empty())
	  {
	    if (sample.type == SAMPLE_LIVE)
	      plotter = (o.plotter ? o.plotter->plotter() : "unknown");
	    else
	      plotter = StringFormat("%1/rootobj/%2/%3%4")
			.arg(o.plotter ? o.plotter->plotter() : "unknown")
			.arg(sampleTypeLabel[sample.type])
			.arg(sample.runnr)
			.arg(sample.dataset);
	    info["MEs"] = StringFormat("%1").arg(status[StringAtom(&stree, system)].nleaves);
	    info["processName"] = system;
	  }

	  info[param] = o.data;
	}

        return makeResponse(StringFormat("{'kind':'DQMSummary', 'items':[%1]," 
                            		 " 'qrender':'%2', 'qplot':'%3'}")
			    .arg(summaryToJSON(summary))
			    .arg(plotter)
			    .arg(qplot),
			    sample.type == SAMPLE_LIVE ? 20 : 300, current,
			    services, workspaceName_, workspaces,
			    submenu, sample, filter, reference, rxstr, rxerr,
			    summary.size(), startTime);
      }
    }

private:
  static std::string
  summaryToJSON(const SummaryMap &summary)
    {
      size_t len = 4 * summary.size();
      SummaryMap::const_iterator si, se;
      KeyValueMap::const_iterator ki, ke;
      for (si = summary.begin(), se = summary.end(); si != se; ++si)
      {
	len += 8 * si->second.size();
	for (ki = si->second.begin(), ke = si->second.end(); ki != ke; ++ki)
	  len += ki->first.size() + 2 * ki->second.size();
      }

      std::string result;
      result.reserve(len + 1);
      for (si = summary.begin(), se = summary.end(); si != se; ++si)
      {
	if (! result.empty())
	  result += ", ";
	result += "{";

	bool first = true;
	for (ki = si->second.begin(), ke = si->second.end(); ki != ke; ++ki, first = false)
	{
	  if (! first)
	    result += ", ";
	  result += stringToJSON(ki->first);
	  result += ":";
	  result += stringToJSON(ki->second);
	}
	
	result += "}";
      }

      return result;
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMQualityWorkspace : public VisDQMWorkspace
{
  struct QMapItem
  {
    std::string location;
    std::string name;
    std::string	reportSummary;
    std::string eventTimeStamp;
    uint64_t version;
  };

  typedef std::map<std::string, QMapItem> QMap;
public:
  VisDQMQualityWorkspace(py::object gui, const std::string &name)
    : VisDQMWorkspace(gui, name)
    {}

  virtual ~VisDQMQualityWorkspace(void)
    {}

  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      const std::string &services = serviceListJSON();
      const std::string &workspaces = workspaceListJSON();

      VisDQMSample sample(sessionSample(session));
      std::string filter(py::extract<std::string>(session.get("dqm.filter")));
      std::string reference(py::extract<std::string>(session.get("dqm.reference")));
      std::string submenu(py::extract<std::string>(session.get("dqm.submenu")));
      std::string rxstr(py::extract<std::string>(session.get("dqm.search")));
      std::string rxerr;
      shared_ptr<Regexp> rxsearch;
      makerx(rxstr, rxsearch, rxerr, Regexp::IgnoreCase);

      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	PyReleaseInterpreterLock nogil;

	VisDQMItems		contents;
	StringAtomSet		myobjs; // deliberately empty, not needed
	VisDQMItems::iterator	ci, ce;
	std::string		plotter;
	std::string		label;
	QMap			qmap;
	VisDQMEventNum		current = { "", -1, -1, -1, -1 };
	bool			alarms = false;
	bool			*qalarm = 0;

	if (filter == "all")
	  qalarm = 0;
	else if (filter == "alarms")
	  alarms = true, qalarm = &alarms;
	else if (filter == "nonalarms")
	  alarms = false, qalarm = &alarms;

	buildContents(srclist, contents, myobjs, sample,
		      current, 0, rxsearch.get(), 0, 0, 0);

	for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
	{
	  VisDQMItem &o = *ci->second;
	  size_t pos = o.name.string().find("/EventInfo/reportSummaryMap");
	  if (pos > 0 && pos != std::string::npos && ! isScalarType(o.flags))
	  {
	    size_t start = o.name.string().rfind('/', pos ? pos-1 : 0);
	    if (start == std::string::npos) start = 0;
	    label = o.name.string().substr(start, pos-start);
	    if (rxsearch && rxsearch->search(label) < 0)
	      continue;

	    StringAtom reportSummary(&stree,
				     label + "/EventInfo/reportSummary",
				     StringAtom::TestOnly);

	    StringAtom eventTimeStamp(&stree,
				      label + "/EventInfo/eventTimeStamp",
				      StringAtom::TestOnly);

	    // Alarm filter. Somewhat convoluted as we need to dig out the
	    // report summary value, and use the value known by JavaScript
	    // to check for items passing an alarm filter here.
	    if (qalarm)
	    {
	      VisDQMItems::iterator pos = contents.find(reportSummary);
	      if (pos != ce && isRealType(pos->second->flags))
	      {
		double value = atof(pos->second->data.c_str());
		if ((value >= 0 && value < 0.95) != alarms)
		  continue;
	      }
	    }

	    QMapItem &i = qmap[label];
	    if (sample.type == SAMPLE_LIVE)
	      i.location = (o.plotter ? o.plotter->plotter() : "unknown");
	    else
	      i.location = StringFormat("%1/rootobj/%2/%3%4")
			   .arg(o.plotter ? o.plotter->plotter() : "unknown")
			   .arg(sampleTypeLabel[sample.type])
			   .arg(sample.runnr)
			   .arg(sample.dataset);
	    
	    i.name = o.name.string();
	    i.version = o.version;
	    i.reportSummary.clear();
	    i.eventTimeStamp.clear();

	    VisDQMItems::iterator other;
	    if ((other = contents.find(reportSummary)) != ce
		&& isRealType(other->second->flags))
	      i.reportSummary = other->second->data;

	    if ((other = contents.find(eventTimeStamp)) != ce
		&& isRealType(other->second->flags))
	      i.eventTimeStamp = other->second->data;
	  }
	}

        return makeResponse(StringFormat("{'kind':'DQMQuality', 'items':[%1]}")
			    .arg(qmapToJSON(qmap)),
			    sample.type == SAMPLE_LIVE ? 20 : 300, current,
			    services, workspaceName_, workspaces,
			    submenu, sample, filter, reference, rxstr, rxerr,
			    qmap.size(), startTime);
      }
    }

private:
  static std::string
  qmapToJSON(const QMap &qmap)
    {
      size_t len = 30 * qmap.size();
      QMap::const_iterator i, e;
      for (i = qmap.begin(), e = qmap.end(); i != e; ++i)
	len += 100 + 2 * (i->first.size()
			  + i->second.location.size()
			  + i->second.name.size()
			  + i->second.reportSummary.size()
			  + i->second.eventTimeStamp.size());

      std::string result;
      result.reserve(len);
      for (i = qmap.begin(), e = qmap.end(); i != e; ++i)
      {
	char buf[32];
	__extension__
	  sprintf(buf, "%ju", (uintmax_t) i->second.version);
	if (! result.empty())
	  result += ", ";
	result += "{'label':";
	result += stringToJSON(i->first);
	result += ",'version':";
	result += buf;
	result += ",'name':";
	result += stringToJSON(i->second.name);
	result += ",'location':";
	result += stringToJSON(i->second.location);
	result += ",'reportSummary':";
	result += stringToJSON(i->second.reportSummary);
	result += ",'eventTimeStamp':";
	result += stringToJSON(i->second.eventTimeStamp);
	result += "}";
      }

      return result;
    }
};


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMSampleWorkspace : public VisDQMWorkspace
{
public:
  VisDQMSampleWorkspace(py::object gui, const std::string &name)
    : VisDQMWorkspace(gui, name)
    {}

  virtual ~VisDQMSampleWorkspace(void)
    {}

  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      VisDQMSample sample(sessionSample(session));
      std::string vary(py::extract<std::string>(session.get("dqm.sample.vary")));
      std::string order(py::extract<std::string>(session.get("dqm.sample.order")));
      std::string pat(py::extract<std::string>(session.get("dqm.sample.pattern")));

      // If the current sample is live, force varying off, otherwise
      // it's too hard for users to find anything else than the live.
      if (sample.type == SAMPLE_LIVE)
	vary = "any";

      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	typedef std::set<std::string> MatchSet;
	typedef std::list< std::pair<std::string, shared_ptr<Regexp> > > RXList;

	PyReleaseInterpreterLock nogil;
	VisDQMSamples		samples;
	VisDQMSamples		final;
	StringList		pats;
	StringList		runlist;
	RXList			rxlist;
	std::string		rxerr;
	MatchSet		uqpats;
        bool			varyAny = (vary == "any");
	bool			varyRun = (vary == "run");
	bool			varyDataset = (vary == "dataset");
	std::string		samplebase;

	if (sample.type == SAMPLE_OFFLINE_RELVAL)
	  samplebase = StringOps::replace(sample.dataset, RX_CMSSW_VERSION, "");

	pats = StringOps::split(pat, Regexp::rxwhite, StringOps::TrimEmpty);
	runlist.reserve(pats.size());
	uqpats.insert(pats.begin(), pats.end());

	// Process space-separated patterns.  Split pure numbers to
	// run number criteria, and others to regexp criteria.  The
	// run number requires a prefix match on the number, not an
	// exact match, and we do the prefix comparison as strings.
	for (size_t i = 0, e = pats.size(); i != e; ++i)
	{
	  errno = 0;
	  char *p = 0;
	  strtol(pats[i].c_str(), &p, 10);
	  if (p && ! *p && ! errno)
	    runlist.push_back(pats[i]);

	  shared_ptr<Regexp> rx;
	  makerx(pats[i], rx, rxerr, Regexp::IgnoreCase);
	  if (rx)
	    rxlist.push_back(RXList::value_type(pats[i], rx));
	}

	// Request samples from backends.
	for (size_t i = 0, e = srclist.size(); i != e; ++i)
	  srclist[i]->samples(samples);

	// Filter out samples which do not pass search criteria.
	// Require all pattern components to match at least one
	// criteria (run number or something in the dataset name).
        final.reserve(samples.size());
	for (size_t i = 0, e = samples.size(); i != e; ++i)
	{
	  MatchSet matched;
	  StringList::iterator vi, ve;
	  StringList::iterator rli, rle;
	  RXList::iterator rxi, rxe;
	  VisDQMSample &s = samples[i];

	  if (varyAny)
	    /* keep all */;
          else if (s.type == SAMPLE_LIVE)
	    /* always provide link to live sample */;
	  else if (varyDataset)
	  {
	    if (sample.type == SAMPLE_OFFLINE_RELVAL)
	    {
	      // current sample has no version set, search dataset name.
	      if (s.type != sample.type
		  || sample.dataset.find(s.version) == std::string::npos)
	        continue;
	    }
	    else if (s.runnr != sample.runnr || s.version != sample.version)
	      continue;
          }
          else if (varyRun)
          {
	    if (sample.type == SAMPLE_OFFLINE_RELVAL)
	    {
	      if (s.type != sample.type
		  || samplebase != StringOps::replace(s.dataset, RX_CMSSW_VERSION, ""))
		continue;
	    }
	    else if (s.dataset != sample.dataset)
	      continue;
	  }
	  else
	    continue;

	  for (rli = runlist.begin(), rle = runlist.end(); rli != rle; ++rli)
	  {
	    char buf[32];
	    sprintf(buf, "%ld", s.runnr);
	    if (! strncmp(buf, rli->c_str(), rli->size()))
	      matched.insert(*rli);
	  }

	  for (rxi = rxlist.begin(), rxe = rxlist.end(); rxi != rxe; ++rxi)
	    if (rxi->second->search(s.dataset) >= 0
		|| rxi->second->search(sampleTypeLabel[s.type]) >= 0)
	      matched.insert(rxi->first);

	  if (matched.size() == uqpats.size())
	    final.push_back(samples[i]);
	}

        std::sort(final.begin(), final.end(),
		   (order == "run" ? orderSamplesByRun
		    : orderSamplesByDataset));

	StringFormat result
	  = StringFormat("([{'kind':'AutoUpdate', 'interval':%1, 'stamp':%2, 'serverTime':%8},"
			 "{'kind':'DQMSample', 'vary':%3, 'order':%4, 'search':%5, 'current':%6,"
			 " 'items':[%7]}])")
	    .arg(sample.type == SAMPLE_LIVE ? 20 : 300)
	    .arg(guiTimeStamp_, 0, 'f')
	    .arg(stringToJSON(vary))
	    .arg(stringToJSON(order))
	    .arg(stringToJSON(pat))
	    .arg(sampleToJSON(sample))
	    .arg(samplesToJSON(final));
	return result.arg((Time::current() - startTime).ns() * 1e-6, 0, 'f');
      }
    }

private:
  static std::string
  samplesToJSON(VisDQMSamples &samples)
    {
      std::string result;

      size_t size = 0;
      VisDQMSamples::iterator i, e;
      for (i = samples.begin(), e = samples.end(); i != e; ++i)
	size += 100 + i->dataset.size();
      result.reserve(size);

      const char *comma = "";
      VisDQMSampleType last = SAMPLE_LIVE;
      for (i = samples.begin(), e = samples.end(); i != e; ++i)
      {
	if (result.empty() || i->type != last)
	{
	  last = i->type;
	  if (! result.empty())
	    result += "]}, ";
	  result += "{'type':'";
	  result += sampleTypeLabel[i->type];
	  result += "', 'items':[";
	  comma = "";
	}

	result += comma;
	sampleToJSON(*i, result);
	comma = ", ";
      }

      if (! result.empty())
	result += "]}";

      return result;
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMLayoutTools
{
protected:
  static std::string
  axisToJSON(const VisDQMAxisOptions &axis)
    {
      return StringFormat("{'min':%1, 'max':%2, 'type':%3}")
	.arg(stringToJSON(axis.min, true))
	.arg(stringToJSON(axis.max, true))
	.arg(stringToJSON(axis.type));
    }

  static std::string
  layoutToJSON(const VisDQMShownItem &x, const VisDQMSample &sample)
    {
      char vbuf[64]; __extension__ sprintf(vbuf, "%ju", (uintmax_t) x.version);

      std::string location;
      if (sample.type == SAMPLE_LIVE)
	location = (x.plotter ? x.plotter->plotter() : "unknown");
      else
	location = StringFormat("%1/%2/%3/%4%5")
		   .arg(x.plotter ? x.plotter->plotter() : "unknown")
		   .arg(x.type)
		   .arg(sampleTypeLabel[sample.type])
		   .arg(sample.runnr)
		   .arg(sample.dataset);

      return StringFormat("{'name':%1, 'desc':%2, 'location':\"%3\","
			  " 'version':\"%4\", 'alarm':%5, 'live':%6,"
			  " 'xaxis':%7, 'yaxis':%8, 'zaxis':%9,"
			  " 'drawopts':%10, 'withref':%11}")
	.arg(stringToJSON(x.name.string(), true))
	.arg(stringToJSON(x.desc, true))
	.arg(location)
	.arg(vbuf)
	.arg(x.nalarm)
	.arg(x.nlive)
	.arg(axisToJSON(x.drawopts.xaxis))
	.arg(axisToJSON(x.drawopts.yaxis))
	.arg(axisToJSON(x.drawopts.zaxis))
	.arg(stringToJSON(x.drawopts.drawopts))
	.arg(stringToJSON(x.drawopts.withref));
    }

  static std::string
  layoutToJSON(const VisDQMShownRow &layout, const VisDQMSample &sample)
    {
      VisDQMShownItems::const_iterator i, e;
      std::string result;
      result.reserve(2048);
      result += '[';
      for (size_t i = 0, e = layout.columns.size(); i != e; ++i)
      {
	if (i > 0)
	  result += ", ";
	result += layoutToJSON(*layout.columns[i], sample);
      }
      result += ']';
      return result;
    }

  static std::string
  layoutToJSON(const VisDQMShownRows &layout, const VisDQMSample &sample)
    {
      VisDQMShownRows::const_iterator i, e;
      std::string result;
      result.reserve(2048);
      result += '[';
      for (size_t i = 0, e = layout.size(); i != e; ++i)
      {
	if (i > 0)
	  result += ", ";
	result += layoutToJSON(*layout[i], sample);
      }
      result += ']';
      return result;
    }

  static std::string
  shownToJSON(const VisDQMItems &contents,
	      const VisDQMStatusMap &status,
	      const VisDQMDrawOptionMap &drawopts,
	      const StringAtom &path,
	      const VisDQMSample &sample,
	      const StringAtomSet &shown)
    {
      std::string prefix;
      if (! path.string().empty())
      {
	prefix.reserve(path.size() + 2);
	prefix += path.string();
	prefix += '/';
      }

      StringAtomList names;
      StringAtomSet subdirs;
      VisDQMItems::const_iterator ci, ce;
      for (ci = contents.begin(), ce = contents.end(); ci != ce; ++ci)
      {
	size_t slash;
	const std::string &name = ci->first.string();
	if (name.size() > prefix.size()
	    && name.compare(0, prefix.size(), prefix) == 0
	    && (slash = name.find('/', prefix.size())) != std::string::npos)
	  subdirs.insert(StringAtom(&stree, std::string(name, 0, slash)));
      }

      std::string result;
      result.reserve(102400);
      result += '[';

      names.insert(names.end(), subdirs.begin(), subdirs.end());
      std::sort(names.begin(), names.end(), natcmpstratom);
      for (size_t i = 0, e = names.size(); i < e; ++i)
      {
	if (result.size() > 1)
	  result += ", ";
	
	result += shownDirToJSON(contents, status, names[i]);
      }
      
      names.clear();
      names.insert(names.end(), shown.begin(), shown.end());
      std::sort(names.begin(), names.end(), natcmpstratom);
      for (size_t i = 0, e = names.size(); i < e; ++i)
      {
	if (result.size() > 1)
	  result += ", ";
	result += shownObjToJSON(contents, status, drawopts, sample, names[i]);
      }

      result += ']';
      return result;
    }

  static std::string
  shownDirToJSON(const VisDQMItems &contents,
		 const VisDQMStatusMap &status,
		 const StringAtom &x)
    {
      assert(status.count(x));
      const VisDQMStatus &info = status.find(x)->second;
      return StringFormat("{'name':%1, 'dir':{'nalarm':%2, 'nlive':%3, 'nleaves':%4}}")
	.arg(stringToJSON(x.string(), true))
	.arg(info.nalarm)
	.arg(info.nlive)
	.arg(info.nleaves);
    }

  static std::string
  shownObjToJSON(const VisDQMItems &contents,
		 const VisDQMStatusMap &status,
		 const VisDQMDrawOptionMap &drawopts,
		 const VisDQMSample &sample,
		 const StringAtom &x)
    {
      VisDQMShownRows layout;
      layoutOne(layout, contents, status, drawopts, x);
      return StringFormat("{'name':%1, 'items':%2}")
	.arg(stringToJSON(x.string()))
	.arg(layoutToJSON(layout, sample));
    }

  static void
  copyAxisOptions(VisDQMAxisOptions &to, const VisDQMAxisOptions &from)
    {
      if (! from.min.empty())
	to.min = from.min;

      if (! from.max.empty())
	to.max = from.max;

      if (! from.type.empty())
	to.type = from.type;
    }

  static void
  copyDrawOptions(VisDQMDrawOptions *to, const VisDQMDrawOptions *from)
    {
      if (! from)
	return;

      if (! from->withref.empty())
	to->withref = from->withref;

      if (! from->drawopts.empty())
	to->drawopts = from->drawopts;

      copyAxisOptions(to->xaxis, from->xaxis);
      copyAxisOptions(to->yaxis, from->yaxis);
      copyAxisOptions(to->zaxis, from->zaxis);
    }

  static void
  setDrawOptions(VisDQMDrawOptions *to,
		 const VisDQMDrawOptions *opt1 = 0,
		 const VisDQMDrawOptions *opt2 = 0)
    {
      to->withref = "no";
      to->drawopts = "";
      to->xaxis.type = "lin";
      to->yaxis.type = "lin";
      to->zaxis.type = "lin";
      copyDrawOptions(to, opt1);
      copyDrawOptions(to, opt2);
    }

  // Build a layout on this node, 1x1 if it's not a layout otherwise.
  static void
  layoutOne(VisDQMShownRows &layout,
	    const VisDQMItems &contents,
	    const VisDQMStatusMap &status,
	    const VisDQMDrawOptionMap &drawopts,
	    const StringAtom &path)
    {
      VisDQMItems::const_iterator          objpos = contents.find(path);
      VisDQMStatusMap::const_iterator      statuspos = status.find(path);

      assert(objpos != contents.end());
      assert(objpos->second->name == path);
      assert(statuspos != status.end());

      const VisDQMItem &obj = *objpos->second;
      const VisDQMStatus &info = statuspos->second;

      if (! obj.layout)
      {
	VisDQMDrawOptionMap::const_iterator drawpos = drawopts.find(path);
	const VisDQMDrawOptions *xopts = 0;
	if (drawpos != drawopts.end())
	  xopts = &drawpos->second;

	shared_ptr<VisDQMShownItem> item(new VisDQMShownItem);
	item->version = obj.version;
	item->nalarm = info.nalarm;
	item->nlive = info.nlive;
	item->name = obj.name;
	item->plotter = obj.plotter;
	item->type = isScalarType(obj.flags) ? "scalar" : "rootobj";
	setDrawOptions(&item->drawopts, xopts);
	shared_ptr<VisDQMShownRow> row(new VisDQMShownRow);
	row->columns.push_back(item);
	layout.push_back(row);
      }
      else
      {
	const VisDQMLayoutRows &rows = *obj.layout;
	layout.reserve(rows.size());

	for (size_t nrow = 0; nrow < rows.size(); ++nrow)
	{
	  const VisDQMLayoutRow &row = *rows[nrow];
	  shared_ptr<VisDQMShownRow> srow(new VisDQMShownRow);
	  srow->columns.reserve(row.columns.size());
	  layout.push_back(srow);

	  for (size_t ncol = 0; ncol < row.columns.size(); ++ncol)
	  {
	    const VisDQMDrawOptions *xopts = 0;
	    const VisDQMDrawOptions *lopts = 0;
	    const VisDQMLayoutItem &col = *row.columns[ncol];
	    shared_ptr<VisDQMShownItem> scol(new VisDQMShownItem);

	    if ((objpos = contents.find(col.path)) != contents.end())
	    {
	      statuspos = status.find(col.path);
	      assert(statuspos != status.end());

	      const VisDQMItem &lobj = *objpos->second;
	      const VisDQMStatus &linfo = statuspos->second;
	      assert(lobj.name == col.path);

	      scol->version = lobj.version;
	      scol->nalarm = linfo.nalarm;
	      scol->nlive = linfo.nlive;
	      scol->name = col.path;
	      scol->desc = col.desc;
	      scol->plotter = lobj.plotter;
	      scol->type = isScalarType(lobj.flags) ? "scalar" : "rootobj";
	      lopts = &col.drawopts;

	      VisDQMDrawOptionMap::const_iterator drawpos = drawopts.find(col.path);
	      if (drawpos != drawopts.end())
		xopts = &drawpos->second;
	    }
	    else
	    {
	      scol->version = 0;
	      scol->nalarm = 0;
	      scol->nlive = 1;
	      scol->name = col.path;
	      scol->desc = col.desc;
	      scol->plotter = 0;
	      scol->type = "none";
	    }

	    setDrawOptions(&scol->drawopts, lopts, xopts);
	    srow->columns.push_back(scol);
	  }
	}
      }
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMContentWorkspace : public VisDQMWorkspace, public VisDQMLayoutTools
{
  shared_ptr<Regexp> rxmatch_;
  shared_ptr<Regexp> rxlayout_;
public:
  VisDQMContentWorkspace(py::object gui,
			 const std::string &name,
			 const std::string &match)
    : VisDQMWorkspace(gui, name),
      rxmatch_(rx(match))
    {
      if (! match.empty() && match != "^")
	rxlayout_ = rx("(" + match + "Layouts)/.*");
    }

  virtual ~VisDQMContentWorkspace(void)
    {}

  // Helper function to present session state.  Returns a JSON object
  // representing current menu and canvas state.
  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      const std::string &services = serviceListJSON();
      const std::string &workspaces = workspaceListJSON();

      VisDQMSample  sample       (sessionSample(session));
      bool	    helpshow	 (py::extract<bool>(session.get("dqm.panel.help.show", true)));
      int	    helpx	 (py::extract<int>(session.get("dqm.panel.help.x", -1)));
      int	    helpy	 (py::extract<int>(session.get("dqm.panel.help.y", -1)));
      bool	    customshow	 (py::extract<bool>(session.get("dqm.panel.custom.show", false)));
      int	    customx	 (py::extract<int>(session.get("dqm.panel.custom.x", -1)));
      int	    customy	 (py::extract<int>(session.get("dqm.panel.custom.y", -1)));
      std::string   root         (workspaceParam<std::string>(session, "dqm.root", ""));
      std::string   focus        (workspaceParam<std::string>(session, "dqm.focus", ""));  // None
      std::string   filter       (py::extract<std::string>(session.get("dqm.filter")));
      std::string   reference    (py::extract<std::string>(session.get("dqm.reference")));
      std::string   submenu      (py::extract<std::string>(session.get("dqm.submenu")));
      std::string   size	 (py::extract<std::string>(session.get("dqm.size")));
      std::string   rxstr        (py::extract<std::string>(session.get("dqm.search")));
      py::dict      pydrawopts   (workspaceParam<py::dict>(session, "dqm.drawopts", py::dict()));
      py::tuple     myobjst      (workspaceParam<py::tuple>(session, "dqm.myobjs", py::tuple()));
      StringAtomSet myobjs;

      for (py::stl_input_iterator<std::string> e, i(myobjst); i != e; ++i)
        myobjs.insert(StringAtom(&stree, *i));

      VisDQMDrawOptionMap drawopts;
      for (py::stl_input_iterator<std::string> i(pydrawopts), e; i != e; ++i)
	translateDrawOptions(py::extract<py::dict>(pydrawopts.get(*i)),
			     drawopts[StringAtom(&stree, *i)]);

      std::string rxerr;
      shared_ptr<Regexp> rxsearch;
      makerx(rxstr, rxsearch, rxerr, Regexp::IgnoreCase);
      
      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	PyReleaseInterpreterLock nogil;

	VisDQMItems		contents;
	StringAtomSet		shown;
	VisDQMItems::iterator	ci, ce;
	VisDQMStatusMap		status;
	std::string		data;
	std::string		layoutroot;
	VisDQMEventNum		current = { "", -1, -1, -1, -1 };
	bool			alarms = false;
	bool			*qalarm = 0;

	if (filter == "all")
	  qalarm = 0;
	else if (filter == "alarms")
	  alarms = true, qalarm = &alarms;
	else if (filter == "nonalarms")
	  alarms = false, qalarm = &alarms;

	buildContents(srclist, contents, myobjs, sample, current,
		      rxmatch_.get(), rxsearch.get(), qalarm,
		      &layoutroot, rxlayout_.get());
	buildShown(contents, shown, root);
	updateStatus(contents, status);

	if (! status.count(StringAtom(&stree, root)))
	  root = ""; // FIXME: #36093
	if (! focus.empty() && ! status.count(StringAtom(&stree, focus)))
	  focus = ""; // None

        return makeResponse(StringFormat("{'kind':'DQMCanvas', 'items':%1,"
					 " 'root':%2, 'focus':%3, 'size':'%4',"
			    		 " 'help':{'show':%5, 'x':%6, 'y':%7},"
					 " 'custom':{'show':%8, 'x':%9, 'y':%10},"
					 " 'reference':'%11', 'layoutroot':%12}")
			    .arg(shownToJSON(contents, status, drawopts,
					     StringAtom(&stree, root),
					     sample, shown))
			    .arg(stringToJSON(root))
			    .arg(stringToJSON(focus, true))
			    .arg(size)
			    .arg(int(helpshow)).arg(helpx).arg(helpy)
			    .arg(int(customshow)).arg(customx).arg(customy)
			    .arg(reference).arg(stringToJSON(layoutroot)),
			    sample.type == SAMPLE_LIVE ? 20 : 300, current,
			    services, workspaceName_, workspaces,
			    submenu, sample, filter, reference, rxstr, rxerr,
			    contents.size(), startTime);
      }
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
class VisDQMPlayWorkspace : public VisDQMWorkspace, public VisDQMLayoutTools
{
public:
  VisDQMPlayWorkspace(py::object gui, const std::string &name)
    : VisDQMWorkspace(gui, name)
    {}

  virtual ~VisDQMPlayWorkspace(void)
    {}

  // Helper function to present session state.  Returns a JSON object
  // representing current menu and canvas state.
  virtual std::string
  state(py::dict session)
    {
      Time startTime = Time::current();
      gui_.attr("_noResponseCaching")();
      std::vector<VisDQMSource *> srclist;
      sources(srclist);

      const std::string &services = serviceListJSON();
      const std::string &workspaces = workspaceListJSON();

      VisDQMSample  sample       (sessionSample(session));
      std::string   workspace    (py::extract<std::string>(session.get("dqm.play.prevws")));
      std::string   root         (workspaceParamOther<std::string>(session, "dqm.root", "", workspace));
      std::string   filter       (py::extract<std::string>(session.get("dqm.filter")));
      std::string   reference    (py::extract<std::string>(session.get("dqm.reference")));
      std::string   submenu      (py::extract<std::string>(session.get("dqm.submenu")));
      std::string   rxstr        (py::extract<std::string>(session.get("dqm.search")));
      int           playpos      (py::extract<int>(session.get("dqm.play.pos")));
      int           playinterval (py::extract<int>(session.get("dqm.play.interval")));
      py::dict      pydrawopts   (workspaceParamOther<py::dict>(session, "dqm.drawopts", py::dict(), workspace));
      py::tuple     myobjst      (workspaceParamOther<py::tuple>(session, "dqm.myobjs", py::tuple(), workspace));
      StringAtomSet myobjs;

      for (py::stl_input_iterator<std::string> e, i (myobjst); i != e; ++i)
        myobjs.insert(StringAtom(&stree, *i));

      VisDQMDrawOptionMap drawopts;
      for (py::stl_input_iterator<std::string> i(pydrawopts), e; i != e; ++i)
	translateDrawOptions(py::extract<py::dict>(pydrawopts.get(*i)),
			     drawopts[StringAtom(&stree, *i)]);

      std::string rxerr;
      shared_ptr<Regexp> rxsearch;
      makerx(rxstr, rxsearch, rxerr, Regexp::IgnoreCase);

      // Give sources pre-scan warning so they can do python stuff.
      for (size_t i = 0, e = srclist.size(); i != e; ++i)
	srclist[i]->prescan();

      // Now do the hard stuff, out of python.
      {
	PyReleaseInterpreterLock nogil;

	VisDQMItems		contents;
	StringAtomSet		shown;
	VisDQMItems::iterator	ci, ce;
	VisDQMStatusMap		status;
	std::string		data;
	VisDQMEventNum		current = { "", -1, -1, -1, -1 };
	bool			alarms = false;
	bool			*qalarm = 0;

	if (filter == "all")
	  qalarm = 0;
	else if (filter == "alarms")
	  alarms = true, qalarm = &alarms;
	else if (filter == "nonalarms")
	  alarms = false, qalarm = &alarms;

	buildContents(srclist, contents, myobjs, sample,
		      current, 0, rxsearch.get(), qalarm, 0, 0);
	buildShown(contents, shown, root);
	updateStatus(contents, status);

	if (! status.count(StringAtom(&stree, root)))
	  root = ""; // FIXME: #36093

	if (shown.size())
	  playpos = std::min(playpos, int(shown.size())-1);
	else
	  playpos = 0;

        return makeResponse(StringFormat("{'kind':'DQMPlay', 'items':%1,"
					 " 'interval':%2, 'pos':%3, 'max':%4}")
			    .arg(shownToJSON(contents, status, drawopts,
					     StringAtom(&stree, root),
					     sample, shown))
			    .arg(playinterval)
			    .arg(playpos)
			    .arg(shown.size()),
			    300, current, services, workspace, workspaces,
			    submenu, sample, filter, reference, rxstr, rxerr,
			    contents.size(), startTime);
      }
    }
};

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
#define EXC_TRANSLATE(n,t,what)					\
  struct translate ## n { static void translate (const t &e)	\
    { PyErr_SetString(PyExc_RuntimeError, e. what); } };	\
  py::register_exception_translator< t >			\
    (&translate ## n :: translate)

template<class A, class B>
struct pair_to_tuple  
{
  typedef pair_to_tuple<A,B> converter;
  typedef std::pair<A,B>     ctype;

  static PyObject* convert(ctype const& v)  
    {
      return py::incref(py::make_tuple(v.first,v.second).ptr());
    }

  static void register_to_python()  
    {
      py::to_python_converter<ctype,converter>();
    }
};

BOOST_PYTHON_MODULE(libVisDQMServer) {
  DebugAids::failHook(&onAssertFail);

  // Re-capture signals from ROOT after ROOT has initialised.
  ROOT::GetROOT();
  for (int sig = 1; sig < _NSIG; ++sig)
    if (sig == SIGPIPE)
      Signal::ignore(sig);
    else
      Signal::revert(sig);

  Signal::handleFatal(0, IOFD_INVALID, 0, 0,
		      (Signal::FATAL_DEFAULT
		       & ~(Signal::FATAL_ON_INT
			   | Signal::FATAL_ON_QUIT
			   | Signal::FATAL_DUMP_CORE)));

  // Initialise locks.
  pthread_rwlock_init(&parentslock, 0);

  // Register types and conversions.
  pair_to_tuple<std::string,std::string>::register_to_python();
  py::enum_<VisDQMObjType>("objtype")
    .value(objTypeLabel[OBJ_SCALAR],  OBJ_SCALAR)
    .value(objTypeLabel[OBJ_ROOTOBJ], OBJ_ROOTOBJ);
  py::enum_<VisDQMSampleType>("sampletype")
    .value(sampleTypeLabel[SAMPLE_LIVE],           SAMPLE_LIVE)
    .value(sampleTypeLabel[SAMPLE_ONLINE_DATA],    SAMPLE_ONLINE_DATA)
    .value(sampleTypeLabel[SAMPLE_OFFLINE_DATA],   SAMPLE_OFFLINE_DATA)
    .value(sampleTypeLabel[SAMPLE_OFFLINE_RELVAL], SAMPLE_OFFLINE_RELVAL)
    .value(sampleTypeLabel[SAMPLE_OFFLINE_MC],     SAMPLE_OFFLINE_MC);

  // Register methods.
  py::class_<VisDQMSource, shared_ptr<VisDQMSource>, boost::noncopyable>
    ("DQMSource", py::no_init)
    .def("prepareSession", &VisDQMSource::prepareSession);

  py::class_<VisDQMUnknownSource, shared_ptr<VisDQMUnknownSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMUnknownSource", py::init<>())
    .add_property("plothook", &VisDQMUnknownSource::plotter);

  py::class_<VisDQMLiveSource, shared_ptr<VisDQMLiveSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMLiveSource", py::init<py::object, py::dict>())
    .add_property("plothook", &VisDQMLiveSource::plotter)
    .def("_data", &VisDQMLiveSource::data)
    .def("_stop", &VisDQMLiveSource::stop);

  py::class_<VisDQMArchiveSource, shared_ptr<VisDQMArchiveSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMArchiveSource", py::init<py::object, py::dict>())
    .add_property("plothook", &VisDQMArchiveSource::plotter)
    .def("_data", &VisDQMArchiveSource::data)
    .def("_stop", &VisDQMArchiveSource::stop);

  py::class_<VisDQMLayoutSource, shared_ptr<VisDQMLayoutSource>,
    	     py::bases<VisDQMSource>, boost::noncopyable>
    ("DQMLayoutSource", py::init<py::list>());

  py::class_<VisDQMWorkspace, shared_ptr<VisDQMWorkspace>, boost::noncopyable>
    ("DQMWorkspace", py::no_init)
    .def("_profilesnap", &VisDQMWorkspace::profilesnap);

  py::class_<VisDQMSampleWorkspace, shared_ptr<VisDQMSampleWorkspace>,
	     py::bases<VisDQMWorkspace>, boost::noncopyable>
    ("DQMSampleWorkspace", py::init<py::object, std::string>())
    .def("_state", &VisDQMSampleWorkspace::state);

  py::class_<VisDQMSummaryWorkspace, shared_ptr<VisDQMSummaryWorkspace>,
	     py::bases<VisDQMWorkspace>, boost::noncopyable>
    ("DQMSummaryWorkspace", py::init<py::object, std::string>())
    .def("_state", &VisDQMSummaryWorkspace::state);

  py::class_<VisDQMQualityWorkspace, shared_ptr<VisDQMQualityWorkspace>,
	     py::bases<VisDQMWorkspace>, boost::noncopyable>
    ("DQMQualityWorkspace", py::init<py::object, std::string>())
    .def("_state", &VisDQMQualityWorkspace::state);

  py::class_<VisDQMContentWorkspace, shared_ptr<VisDQMContentWorkspace>,
	     py::bases<VisDQMWorkspace>, boost::noncopyable>
    ("DQMContentWorkspace", py::init<py::object, std::string, std::string>())
    .def("_state", &VisDQMContentWorkspace::state);

  py::class_<VisDQMPlayWorkspace, shared_ptr<VisDQMPlayWorkspace>,
	     py::bases<VisDQMWorkspace>, boost::noncopyable>
    ("DQMPlayWorkspace", py::init<py::object, std::string>())
    .def("_state", &VisDQMPlayWorkspace::state);

  EXC_TRANSLATE(StdError, std::exception, what());
  EXC_TRANSLATE(LATError, lat::Error, explain().c_str());
}
