#ifndef VISMONITORING_DQMSERVER_VISDQMTOOLS_H
# define VISMONITORING_DQMSERVER_VISDQMTOOLS_H

# include "VisMonitoring/DQMServer/interface/VisDQMFile.h"
# include "VisMonitoring/DQMServer/interface/StringAtom.h"
# include "classlib/utils/DebugAids.h"
# include "classlib/utils/Signal.h"
# include "TBufferFile.h"
# include "TStreamerInfo.h"
# include "TClass.h"
# include <iostream>

/// Options on how we want fatal signals to be handled.
static const int FATAL_OPTS = (lat::Signal::FATAL_DEFAULT
			       & ~(lat::Signal::FATAL_ON_INT
				   | lat::Signal::FATAL_ON_QUIT
				   | lat::Signal::FATAL_DUMP_CORE));

// ----------------------------------------------------------------------
/** Assertion failure handler.  Always abort without any input. */
inline char
onAssertFail(const char *message)
{
  std::cout.flush(); fflush(stdout);
  std::cerr.flush(); fflush(stderr);
  std::cerr << message << "ABORTING\n";
  return 'a';
}

//----------------------------------------------------------------------
/** Extract a string table with top 32-bit key @a idx in DQM file @a f
    into @a tree.  The full string contents are stored as is, without
    a terminating null character, so binary data blobs are also ok. */
inline void
readStrings(StringAtomTree &tree, VisDQMFile *f, uint64_t idx)
{
  void *begin;
  void *end;
  uint64_t key;
  uint64_t hipart;
  uint64_t lopart;

  for (VisDQMFile::ReadHead rdhead(f, idx); ! rdhead.isdone(); rdhead.next())
  {
    rdhead.get(&key, &begin, &end);
    hipart = key & 0xffffffff00000000ull;
    lopart = key & 0x00000000ffffffffull;
    if (hipart != idx)
      break;

    std::string data((const char *) begin, (const char *) end);
    VERIFY(tree.insert(data) == lopart+1);
  }
}

/** Write a string table @a tree into a file at @a wrhead with the top
    32-bits equal to @a idx and low 32-bits the tree key index.  The
    full string contents is written out with a terminating null byte,
    so binary data blobs with embedded nulls are also ok. */
inline void
writeStrings(VisDQMFile::WriteHead &wrhead, StringAtomTree &tree, uint64_t idx)
{
  for (size_t i = 1, e = tree.size(); i < e; ++i)
  {
    void *wrbegin, *wrend;
    const std::string &str = tree.key(i);
    wrhead.allocate(idx + i - 1, str.size(), &wrbegin, &wrend);
    memcpy(wrbegin, str.c_str(), str.size());
  }
}

//----------------------------------------------------------------------
/** Extract the next serialised ROOT object from @a buf.  Returns null
    if there are no more objects in the buffer, or a null pointer was
    serialised at this location. */
inline TObject *
extractNextObject(TBufferFile &buf)
{
  if (buf.Length() == buf.BufferSize())
    return 0;

  buf.InitMap();
  Int_t pos = buf.Length();
  TClass *c = buf.ReadClass();
  buf.SetBufferOffset(pos);
  buf.ResetMap();
  return c ? buf.ReadObject(c) : 0;
}

/** Load ROOT streamer info from a serialised blob @a data.  It is
    assumed the serialised blob was created by #buildStreamerInfo(). */
inline void
loadStreamerInfo(const std::string &data)
{
  TBufferFile buf(TBufferFile::kRead, data.size(), (void *) &data[0], kFALSE);
  buf.Reset();
  while (buf.Length() != buf.BufferSize())
  {
    TObject *obj = extractNextObject(buf);
    if (TStreamerInfo *si = dynamic_cast<TStreamerInfo *>(obj))
    {
      Int_t v = si->GetClassVersion();
      if (TClass *c = TClass::GetClass(si->GetName(), kTRUE))
      {
	c->GetStreamerInfo();
	if (c->GetStreamerInfos()->At(v))
	{
	  DEBUG(2, "skipping already present streamer info version " << v
		<< " for '" << si->GetName() << "'\n");
	  continue;
	}
      }

      si->BuildCheck();
      DEBUG(2, "importing streamer info version " << v
	    << " for '" << si->GetName() << "'\n");
    }
    else if (! obj)
      std::cerr << "warning: unexpected null streamer info\n";
    else
      std::cerr << "warning: unexpected object embedded in streamer info\n";
  }
}

/** Build a streamer info data blob corresponding to the current ROOT
    version into @a data.  This represents the ROOT version of the
    presently running program, not the version of the ROOT objects
    read in: ROOT will convert whatever we read into the version we
    run now.  This function should be called before any ROOT files
    have been opened as file opens may modify the streamer info. */
inline void
buildStreamerInfo(std::string &data)
{
  static const char *classes[] = 
    {
      "TH1F", "TH1S",
      "TH2F", "TH2S",
      "TH3F", "TH3S",
      "TProfile", "TProfile2D",
      0
    };

  TBufferFile buf(TBufferFile::kWrite);
  for (const char **name = classes; *name; ++name)
    buf.WriteObject(TClass::GetClass(*name)->GetStreamerInfo());

  data.resize(buf.Length());
  memcpy(&data[0], buf.Buffer(), buf.Length());
}

#endif // VISMONITORING_DQMSERVER_VISDQMTOOLS_H
