/** The menu on the top of the screen in the DQM GUI.

    DQMHeaderRow shows basic interactive status information and
    messages.  It provides short cut access to the other DQM GUI
    instances in the same "ring", the workspaces, and to the data
    samples by run and by dataset.  It shows the current run, lumi
    section and event numbers and various messages.  It provides
    general options such as a search and an alarm filter.

    Note the "current run number" shown at the top of the menu should
    not be assumed to be the "run number" of the current data sample.
    For one the "current numbers" are pretty-formatted to include
    thousand separators and cannot be compared with numerical values.
    For another the values contain text such as "(None)" when the
    values could not be determined or "(Simulated)" for Monte Carlo
    samples which do not have run numbers.  Lastly, the online live
    does not have an intrinsic run number, and the "current run" may
    be text such as "(Waiting)" if the run number could not be
    determined. */
GUI.Plugin.DQMHeaderRow = new function() {
  /** Reference for this, for use in callbacks where "this" is
      ambiguous or at least not this closure. */
  var _self = this;

  /** Reference to the GUI main object, for making server calls. */
  var _gui = null;

  /** Data from the last update.  Used to fill in the interaction area
      with a menu of the user's choice (data, services or workspaces)
      without having to make another round trip server call. */
  var _data = null;

  /** The DOM element for the main header. */
  var _header = $('header');

  /** The DOM element for the sub-header below the main header, where
      we show three interaction menus: data, services or
      workspaces. */
  var _subheader = $('subheader');

  /** The DOM element for the service main menu.  Points to the 'span'
      element to be filled in with the current service name. */
  var _service = null;

  /** The DOM element for the current run number in the main menu.
      Points to the 'span' element to be filled in with the value. */
  var _runNr = null;

  /** The DOM element for the current lumi number in the main menu.
      Points to the 'span' element to be filled in with the value. */
  var _lumiNr = null;

  /** The DOM element for the current event number in the main menu.
      Points to the 'span' element to be filled in with the value. */
  var _eventNr = null;

  /** The DOM element for the current run start in the main menu.
      Points to the 'span' element to be filled in with the value. */
  var _runStart = null;

  /** The DOM element for the current workspace in the main menu.
      Points to the 'span' element to be filled in with the name. */
  var _workspace = null;

  /** The DOM element for data operations sub-header menu.  Points to
      the 'div' element containing the menu.  The menu is shown by
      setting its style.display to '' and style.display of the other
      menus to 'none'. */
  var _menuData = null;

  /** The DOM element for service choice sub-header menu.  Points to
      the 'div' element containing the menu.  The menu is shown by
      setting its style.display to '' and style.display of the other
      menus to 'none'. */
  var _menuService = null;

  /** The DOM element for workspace choice sub-header menu.  Points to
      the 'div' element containing the menu.  The menu is shown by
      setting its style.display to '' and style.display of the other
      menus to 'none'. */
  var _menuWorkspace = null;

  /** The DOM element for the search input field in the data menu. */
  var _search = null;

  /** The DOM element for the search result in the data menu. */
  var _searchResult = null;

  /** An time-out timer for auto-search.  Started when a key press or
      paste is received on the search field.  Once the timer expires
      will check if the search input field has changed and if so
      issues an incremental server side search. */
  var _searchDelay = null;

  /** The last search input field value seen.  Used by the search
      timer function to determine if a new search is needed. */
  var _searchString = "";

  // ----------------------------------------------------------------
  /** Return list of current workspaces, for use by other modules. */
  this.workspaces = function()
  {
    return _data.workspaces;
  };

  // ----------------------------------------------------------------
  /** Utility method to create a new top level heading with @a label
      as the title above and @a content as the contents below.

      If @a link is non-empty, the whole heading element is made a
      link which will call DQMHeaderRow with the method name and
      parameters specified in link as a text string; a link
      'foo("arg")' would call DQMHeaderRow.foo("arg") when the header
      element is clicked.  As usual with links, the method should
      return @c false to prevent link traversal and should not make
      assumptions about @c this.

      Returns the 'h2' DOM element which contains the @a contents.  */
  this.heading = function(label, content, link)
  {
    var linkattr = (link ? " onclick='return GUI.Plugin.DQMHeaderRow." + link + "'" : "")
    var linkstart = (link ? ("<a href='#'" + linkattr + ">") : "");
    var linkend = (link ? ("<img src='" + ROOTPATH + "/static/down-menu.gif'"
      + "style='margin-left:5px;padding-bottom:1px' width='7' height='7' /></a>") : "");

    var div = document.createElement("div");
    _header.appendChild(div);

    div.className = "heading";
    if (link)
      div.style.cursor = "pointer";

    div.innerHTML = ("<div class='label'>"
		     + linkstart + label + linkend + "</div>"
		     + "<h2" + linkattr + ">" + content + "</h2>");

    return div.lastChild;
  };

  // ----------------------------------------------------------------
  /** Attach this plug-in to the GUI display. Fills the header and
      sub-header document elements with the main and secondary
      menus. */
  this.attach = function(gui)
  {
    _gui = gui;
    _service = this.heading("Service", "<span>_</span>:", "pickMenu(\"service\")").firstChild;
    _workspace = this.heading("Workspace", "<span>_</span><span class='delimiter'>&nbsp;&nbsp;.</span>",
			      "pickMenu(\"workspace\")").firstChild;
    _runNr = this.heading("Run #", "<span>_</span><span class='delimiter'>"
		          + "&nbsp;&nbsp;.</span>", "pickSample(\"run\")").firstChild;
    _lumiNr = this.heading("LS #", "<span>_</span><span class='delimiter'>"
			   + "&nbsp;&nbsp;.</span>").firstChild;
    _eventNr = this.heading("Event #", "<span>_</span><span class='delimiter'>"
			    + "&nbsp;&nbsp;.</span>").firstChild;
    _runStart = this.heading("Run started", "<span>_</span>").firstChild;

    _subheader.appendChild(_menuData = document.createElement("div"));
    _menuData.style.display = "none";

    _subheader.appendChild(_menuService = document.createElement("div"));
    _menuService.style.display = "none";

    _subheader.appendChild(_menuWorkspace = document.createElement("div"));
    _menuWorkspace.style.display = "none";

    _menuData.innerHTML = ""
      + "<span class='item'>Reference <select id='subhead-ref'"
      + " onchange='GUI.Plugin.DQMHeaderRow.setReference(this.value)'>"
      + "<option value='overlay'>Overlay</option>"
      + "<option value='on-side'>On side</option>"
      + "<option value='customise' selected='selected'>Customise</option></select></span>"
      + "<span class='item'>Show <select id='subhead-filter'"
      + " onchange='GUI.Plugin.DQMHeaderRow.setFilter(this.value)'>"
      + "<option value='all' selected='selected'>All</option>"
      + "<option value='alarms'>Alarms</option>"
      + "<option value='nonalarms'>Non-alarms</option></select></span>"
      + "<span class='item'>Search <input id='subhead-search' type='text' style='width:8em' />"
      + "</span><span id='subhead-searchres' style='margin-left:5px'>0 objects</span>"
      + "<span class='item'></span>";

    _search = $('subhead-search')
    _searchResult = $('subhead-searchres');
    _search.onkeyup = _search.paste = function(e) {
      if (_searchDelay)
        clearTimeout(_searchDelay);
      _searchDelay = setTimeout(_self.search, 50);
    };
  };

  /** Detach this plug-in from the GUI display.  Clears the header and
      sub-header document elements from the objects #attach() added
      and clears closure references for garbage collection. */
  this.detach = function()
  {
    // Remove all 'heading' elements we added.
    for (var m, n = _header.firstChild; n; n = m)
    {
      m = n.nextSibling;
      if (n.className == "heading")
	_header.removeChild(n);
    }

    // Clear subheader entirely.
    while (_subheader.firstChild)
      _subheader.removeChild(_subheader.firstChild);

    // Null out our references so objects are freed.
    _service = null;
    _workspace = null;
    _runNr = null;
    _lumiNr = null;
    _eventNr = null;
    _runStart = null;
    _menuData = null;
    _menuService = null;
    _menuWorkspace = null;
    _search = null;
    _searchResult = null;
    _searchString = "";

    // If the search delay is activated, clear it.
    if (_searchDelay)
    {
      clearTimeout(_searchDelay);
      _searchDelay = null;
    }
  };

  // ----------------------------------------------------------------
  /** Response callback for activating a particular sub-header menu.
      Switches the menu in place and lets the server know the current
      choice.  We don't wait for a new server state here as we already
      have all the data needed to show the menu, we just want the
      session to have the right state for the next scheduled
      update. */
  this.pickMenu = function(which)
  {
    _self.displayMenu(which);
    YAHOO.util.Connect.asyncRequest("GET", _url() + "/setSubMenu?n=" + which);
    return false;
  };

  /** Response callback for selecting another sample.  The @a vary
      parameter specifies which part should be varied, and must
      be either "run" or "dataset".  Sends user to a special hidden
      "data selector" workspace where one can choose another data
      sample. */
  this.pickSample = function(vary)
  {
    _gui.makeCall(_url() + "/chooseSample?vary=" + vary);
    return false;
  };

  /** Response callback for changing workspace.  The @a choice
      parameter specifies the label of the workspace to issue a server
      call for.  The server automatically switches the sub-header menu
      back to "data" as a part of the workspace switch. */
  this.setWorkspace = function(choice)
  {
    _gui.makeCall(_url() + "/workspace?name=" + encodeURIComponent(choice));
    return false;
  };

  /** Respone callback for changing the alarm filter.  The @a choice
      parameter specifies the filter choice, one of three hard-coded
      values "all", "alarms" or "nonalarms". */
  this.setFilter = function(choice)
  {
    _gui.makeCall(_url() + "/setFilter?n=" + encodeURIComponent(choice));
    return false;
  };

  /** Respone callback for changing the global reference display.
      The @a choice parameter specifies the reference display choice,
       one of three hard-coded values "overlay", "on-side" or "customise". */
  this.setReference = function(choice)
  {
    _gui.makeCall(_url() + "/setReference?n=" + encodeURIComponent(choice));
    return false;
  };

  /** Response callback triggered 50ms after the last change to the
      search input.  Checks if the search input has changed value and
      if so issues a server side incremental search. */
  this.search = function()
  {
    if (_searchString != _search.value)
    {
      _searchString = _search.value;
      _searchResult.innerHTML = "(searching)";
      _gui.makeCall(_url() + "/setSearch?rx=" + encodeURIComponent(_search.value));
    }
  };

  // ----------------------------------------------------------------
  /** Update this plug-in display.  Updates menu areas and document
      title according to the current state. */
  this.update = function(data)
  {
    _data = data;
    _service.innerHTML = data.service;
    _workspace.innerHTML = data.workspace;
    _runNr.innerHTML = data.run;
    _lumiNr.innerHTML = data.lumi;
    _eventNr.innerHTML = data.event;
    _runStart.innerHTML = "<span title='UTC time'>" + data.runstart + "</span>";

    var title = data.service + " - " + data.workspace + " - " + HEADING;
    if (document.title != title)
      document.title = title;

    data.workspaces.sort(function(a, b) {
      if (a.rank != b.rank)
        return a.rank - b.rank;
      else if (a.title < b.title)
        return -1;
      else if (a.title > b.title)
        return 1;
      else
        return 0;
    });

    this.displayMenu(data.view.show);
  };

  /** Switch to displaying sub-header menu @a which, one of
      "workspace", "service" or "data". */
  this.displayMenu = function(which)
  {
    // Switch the sub-header to the workspace menu.  Update the div to
    // a multi-column table workspaces, colum by category.  Workspaces
    // other than the current one are links to setWorkspace() response
    // callback.  Add a "(Back)" link after the table so the user can
    // go back to data menu without switching workspaces.  Skip
    // workspaces with a negative ranking, these are hidden internal
    // implementation workspaces added automatically by the server.
    if (which == 'workspace')
    {
      var category = null;
      var cellattr = "<td style='border-right:1px dotted #ddd;padding:0 1em 0 ";
      var subhead = "<table id='subhead-workspace' class='item'><tbody valign='top'><tr>";
      for (var c = 0, i = 0, e = _data.workspaces.length; i < e; ++i)
      {
	var item = _data.workspaces[i];
	if (item.rank < 0)
	  continue;

	if (item.category != category)
	{
	  subhead += (category != null ? "</td>" : "")
		     + cellattr
		     + (category != null ? "10px'>" : "0'>")
		     + "<b style='color:#b66'>"
		     + item.category + "</b><br />";
          category = item.category;
	  ++c;
	}

	if (item.title != _data.workspace)
	  subhead += "<a href='#' onclick=\"return GUI.Plugin.DQMHeaderRow"
		     + ".setWorkspace('" + item.label + "')\">" + item.title
		     + "</a>";
	else
	  subhead += item.title;

	if (i < e-1)
          subhead += "<br />";
      }

      subhead += "</td>" + cellattr + "10px;border:none'><a href='#' onclick='"
		 + "return GUI.Plugin.DQMHeaderRow.pickMenu(\"data\")'>"
		 + "(Back)</a></td></tr></tbody></table>";

      if (_menuWorkspace.innerHTML != subhead)
	_menuWorkspace.innerHTML = subhead;

      _menuData.style.display = "none";
      _menuService.style.display = "none";
      _menuWorkspace.style.display = "";
    }

    // Switch the sub-header to the service menu.  Update the div to a
    // list of links to DQM GUI services, except leave the current one
    // as a non-link name.  Add a "(Back)" link after the list so the
    // user can go back to data menu without switching services.  Note
    // that these are real external links, not AJAX call links.
    else if (which == 'service')
    {
      var subhead = "";
      for (var i = 0, e = _data.services.length; i < e; ++i)
      {
	var item = _data.services[i];
	subhead += "<span class='item'>"
		   + (item.title == _data.service ? item.title : 
		      ("<a href='" + item.href + "'>" + item.title + "</a>"))
		   + "</span>";
      }

      subhead += "<span class='item'><a href='#' onclick='"
		 + "return GUI.Plugin.DQMHeaderRow.pickMenu(\"data\")'>"
		 + "(Back)</a></span>";

      if (_menuService.innerHTML != subhead)
	_menuService.innerHTML = subhead;

      _menuData.style.display = "none";
      _menuService.style.display = "";
      _menuWorkspace.style.display = "none";
    }

    // Switch the sub-header to the data menu.  Show the type of data
    // the user is looking at and the name of the dataset as a link to
    // the sample selection menu.  Show alarm and search input fields.
    else if (which == 'data')
    {
      // Update dataset.
      var subhead = "";
      var dslink = "<a href='#' onclick='return GUI.Plugin.DQMHeaderRow.pickSample(\"dataset\")'>";
      var s = _data.view.sample;
      if (s.type == "live")
	dslink += "Live";
      else if (s.type == "online_data")
	dslink += "Online";
      else if (s.type == "offline_data")
	subhead += "Data: ", dslink += _sanitise(s.dataset);
      else if (s.type == "offline_relval")
	subhead += "RelVal: ", dslink += _sanitise(s.dataset);
      else if (s.type == "offline_mc")
	subhead += "MC: ", dslink += _sanitise(s.dataset);
      else
	subhead += "Other: ", dslink += _sanitise(s.dataset);

      subhead += dslink + "</a>";

      if (_menuData.lastChild.innerHTML != subhead)
	_menuData.lastChild.innerHTML = subhead;

      // Update the all/alarm/non-alarm filter.
      var filter = $('subhead-filter');
      if (_data.view.filter == "all")
	filter.selectedIndex = 0;
      else if (_data.view.filter == "alarms")
	filter.selectedIndex = 1;
      else if (_data.view.filter == "nonalarms")
	filter.selectedIndex = 2;

      // Update the overlay/on-side/customise reference switch.
      var ref = $('subhead-ref');
      if (_data.view.reference == "overlay")
	ref.selectedIndex = 0;
      else if (_data.view.reference == "on-side")
	ref.selectedIndex = 1;
      else if (_data.view.reference == "customise")
	ref.selectedIndex = 2;

      // Update the search info.
      var search = _data.view.search;
      var word = (search.pattern == '' ? " object" : " match");
      var plural = (search.pattern == '' ? "s" : "es");
      var searchres = (! search.valid ? "(invalid regexp)"
	: (search.nmatches > 1 ? "(" + search.nmatches + word + plural + ")"
	: search.nmatches == 1 ? "(1 " + word + ")"
	: "(no " + word + plural + ")"));
      if (! _searchDelay && _search.value != search.pattern)
	_search.value = _searchString = search.pattern;
      if (_searchResult.innerHTML != searchres)
	_searchResult.innerHTML = searchres;
      
      _menuData.style.display = "";
      _menuService.style.display = "none";
      _menuWorkspace.style.display = "none";
    }

    // Hide sub-header menu entirely.
    else if (which == 'none')
    {
      _menuData.style.display = "none";
      _menuService.style.display = "none";
      _menuWorkspace.style.display = "none";
    }
  };

  return this;
}();
